"""
The Spectrum class to contain and manipulate spectra.
"""

import warnings
from copy import deepcopy

import astropy.units as u
import numpy as np
import pandas as pd
from astropy.coordinates import ITRS, SkyCoord, SpectralCoord, StokesCoord
from astropy.coordinates.spectral_coordinate import attach_zero_velocities
from astropy.io import registry
from astropy.io.fits.verify import VerifyWarning
from astropy.modeling.fitting import LinearLSQFitter

# from astropy.nddata.ccddata import fits_ccddata_writer
from astropy.table import Table
from astropy.time import Time
from astropy.utils.masked import Masked
from astropy.wcs import WCS, FITSFixedWarning
from ndcube import NDCube
from specutils import Spectrum1D

from dysh.spectra import core

from ..coordinates import (  # is_topocentric,; topocentric_velocity_to_frame,
    KMS,
    Observatory,
    astropy_convenience_frame_names,
    astropy_frame_dict,
    change_ctype,
    get_velocity_in_frame,
    make_target,
    replace_convention,
    sanitize_skycoord,
    veldef_to_convention,
)
from ..log import HistoricalBase, log_call_to_history  # , logger
from ..plot import specplot as sp
from ..util import minimum_string_match
from . import baseline, get_spectral_equivalency

# from astropy.nddata import StdDevUncertainty

# Spectrum attributes to be ignored by Spectrum._copy_attributes
IGNORE_ON_COPY = [
    "_data",
    "_flux",
    "_meta",
    "_mask",
    "_weights",
    "_baseline_model",
    "_plotter",
    "_uncertainty",
    "_unit",
]


class Spectrum(Spectrum1D, HistoricalBase):
    """
    This class contains a spectrum and its attributes. It is built on
    `~specutils.Spectrum1D` with added attributes like baseline model.
    Note that `~specutils.Spectrum1D` can contain multiple spectra but
    we probably will not use that because the restriction that it can
    have only one spectral axis conflicts with slight Doppler shifts.
    See `~specutils.Spectrum1D` for the instantiation arguments.
    """

    @log_call_to_history
    def __init__(self, *args, **kwargs):
        HistoricalBase.__init__(self)
        self._target = kwargs.pop("target", None)
        if self._target is not None:
            self._target = sanitize_skycoord(self._target)
            self._velocity_frame = self._target.frame.name
        else:
            self._velocity_frame = None
        self._observer = kwargs.pop("observer", None)
        Spectrum1D.__init__(self, *args, **kwargs)
        # Try making a target from meta. If it fails, don't worry about it.
        if False:
            if self._target is None:
                try:
                    self._target = make_target(self.meta)
                    self._velocity_frame = self._target.frame.name
                except Exception:
                    pass
        self._spectral_axis._target = self._target
        if self.velocity_convention is None and "VELDEF" in self.meta:
            self._spectral_axis._doppler_convention = veldef_to_convention(self.meta["VELDEF"])
        # if "uncertainty" not in kwargs:
        # self._uncertainty = StdDevUncertainty(np.ones_like(self.data), unit=self.flux.unit)
        # Weights are Non here because can be defined
        # separately from uncertainty by the user. Not really recommended
        # but they can do so. See self.weights()
        # self._weights = None
        self._weights = np.ones(self.flux.shape)
        #  Get observer quantity from observer location.
        if "DATE-OBS" in self.meta:
            self._obstime = Time(self.meta["DATE-OBS"])
        elif "MJD-OBS" in self.meta:
            self._obstime = Time(self.meta["MJD-OBS"])
        else:
            self._obstime = None
        self._spectral_axis._observer = self.observer
        if self._spectral_axis._observer is not None:
            self._velocity_frame = self._spectral_axis._observer.name

        # if mask is not set via the flux input (NaNs in flux or flux.mask),
        # then set the mask to all False == good
        if self.mask is None:
            self._mask = np.full(np.shape(self.flux), False)
        self._baseline_model = None
        self._subtracted = False
        self._normalized = False
        self._exclude_regions = None
        self._include_regions = None  # do we really need this?
        self._plotter = None
        # `self._resolution` is the spectral resolution in channels.
        # This will be 1 for VEGAS, and >1 for the ACS.
        if "FREQRES" in self.meta and "CDELT1" in self.meta:
            self._resolution = self.meta["FREQRES"] / abs(self.meta["CDELT1"])
        else:
            self._resolution = 1

    def _len(self):
        """return the size of the `Spectrum` in the spectral dimension.
        @todo __len__  has unintended consquences, yuck.
        """
        return len(self.frequency)

    @property
    def weights(self):
        """The channel weights of this spectrum"""
        # Cannot take power of NDUncertainty, so convert
        # to a Quanity first.
        # if self._weights is None:
        #    return self._uncertainty.quantity**-2
        return self._weights

    @property
    def exclude_regions(self):
        """The baseline exclusion region(s) of this spectrum"""
        return self._exclude_regions

    ##@todo
    # def exclude_region(self,region):
    # where region is SpectralRegion, channels, velocity, etc.  See core.py baseline method.
    #
    # def region_to_mask():
    #  set spectrum mask to True inside exclude_regions. normally we don't do this for baselining

    @property
    def baseline_model(self):
        """Returns the computed baseline model or None if it has not yet been computed."""
        return self._baseline_model

    @log_call_to_history
    def baseline(self, degree, exclude=None, include=None, **kwargs):
        # fmt: off
        """
        Compute and optionally remove a baseline.
        The code uses `~astropy.modeling.fitting.Fitter` and `~astropy.modeling.polynomial` to compute the baseline.
        See the documentation for those modules for details.
        This method will set the `baseline_model` attribute to the fitted model function which can be evaluated over a domain.

        Note that `include` and `exclude` are mutually exclusive. If both are present, only `include` will be used.

        Parameters
        ----------
        degree : int
            The degree of the polynomial series, a.k.a. baseline order
        exclude : list of 2-tuples of int or `~astropy.units.Quantity`, or `~specutils.SpectralRegion`
            List of region(s) to exclude from the fit.  The tuple(s) represent a range in the form [lower,upper], inclusive.

            Examples:

            One channel-based region: [11,51]

            Two channel-based regions: [(11,51),(99,123)].

            One `~astropy.units.Quantity` region: [110.198*u.GHz,110.204*u.GHz].

            One compound `~specutils.SpectralRegion`: SpectralRegion([(110.198*u.GHz,110.204*u.GHz),(110.196*u.GHz,110.197*u.GHz)]).

            Default: no exclude region

        include : list of 2-tuples of int or `~astropy.units.Quantity`, or `~specutils.SpectralRegion`
            List of region(s) to include in the fit. The tuple(s) represent a range in the form [lower,upper], inclusive.
            See `exclude` for examples.
        model : str
            One of 'polynomial', 'chebyshev', 'legendre', or 'hermite'
            Default: 'chebyshev'
        fitter : `~astropy.modeling.fitting.Fitter`
            The fitter to use. Default: `~astropy.modeling.fitting.LinearLSQFitter` (with `calc_uncertaintes=True`).
            Be careful when choosing a different fitter to be sure it is optimized for this problem.
        remove : bool
            If True, the baseline is removed from the spectrum. Default: False
        """
        # fmt: on
        # @todo: Are exclusion regions OR'd with the existing mask? make that an option?
        kwargs_opts = {
            "remove": False,
            "model": "chebyshev",
            "fitter": LinearLSQFitter(calc_uncertainties=True),
        }
        kwargs_opts.update(kwargs)

        # `include` and `exclude` are mutually exclusive, but we allow `include`
        # if `include` is used, transform it to `exclude`.
        if include != None:
            if exclude != None:
                logger.info(f"Warning: ignoring exclude={exclude}")
            exclude = core.include_to_exclude_spectral_region(include, self)

        self._baseline_model = baseline(self, degree, exclude, **kwargs)

        if kwargs_opts["remove"]:
            s = self.subtract(self._baseline_model(self.spectral_axis))
            self._data = s._data
            self._subtracted = True

    # baseline
    @log_call_to_history
    def undo_baseline(self):
        """
        Undo the most recently computed baseline. If the baseline
        has been subtracted, it will be added back. The `baseline_model`
        attribute is set to None. Exclude regions are untouched.
        """
        if self._baseline_model is None:
            return
        if self._subtracted:
            if self._normalized:
                warnings.warn("Cannot undo previously normalized baseline subtraction")
                return
            s = self.add(self._baseline_model(self.spectral_axis))
            self._data = s._data
            self._baseline_model = None

    def _set_exclude_regions(self, exclude):
        """
        Set the mask for the regions to exclude.

        Parameters
        ----------
        exclude : list of 2-tuples of int or ~astropy.units.Quantity, or ~specutils.SpectralRegion
            List of region(s) to exclude from the fit.  The tuple(s) represent a range in the form [lower,upper], inclusive.
            In channel units.

            Examples: One channel-based region: [11,51],
                      Two channel-based regions: [(11,51),(99,123)].
                      One ~astropy.units.Quantity region: [110.198*u.GHz,110.204*u.GHz].
                      One compound ~specutils.SpectralRegion: SpectralRegion([(110.198*u.GHz,110.204*u.GHz),(110.196*u.GHz,110.197*u.GHz)]).

        """
        pass

    def list_to_spectral_region(self, inlist):
        # @todo utility code to convert a input list of channels or quantities to a spectral region with units of self.spectral_axis.unit.
        # This could go in core.py combine this with _set_exclude_regions
        pass

    def bshow(self):
        """Show the baseline model"""
        print(f"baseline model {self._baseline_model}")

    def plot(self, **kwargs):
        if self._plotter is None:
            self._plotter = sp.SpectrumPlot(self, **kwargs)
        self._plotter.plot(**kwargs)

    @property
    def obstime(self):
        return self._obstime

    @property
    def plotter(self):
        return self._plotter

    def stats(self, roll=0, qac=False):
        """
        Compute some statistics of this `Spectrum`.  The mean, rms, median,
        data minimum and data maximum are calculated.  Note this works
        with slicing, so, e.g.,  `myspectrum[45:153].stats()` will return
        the statistics of the slice.

        Parameters
        ----------
        roll : int
            Return statistics on a 'rolled' array differenced with the
            original array. If there is no correllaton between channels,
            a roll=1 would return an RMS sqrt(2) larger than that of the
            input array. Another advantage of rolled statistics it will
            remove most slow variations, thus RMS/sqrt(2) might be a better
            indication of the underlying RMS.
            Default: 0
        qac : bool
            If set, the returned simple string contains mean,rms,datamin,datamax
            for easier visual regression. Based on some legacy code.
            Default: False

        Returns
        -------
        stats : dict
            Dictionary consisting of (mean,median,rms,datamin,datamax)

        """

        if roll == 0:
            mean = self.mean()
            median = self.median()
            rms = np.nanstd(self.flux)
            dmin = self.min()
            dmax = self.max()
        else:
            d = self[roll:] - self[:-roll]
            mean = d.mean()
            median = d.median()
            rms = np.nanstd(d.flux)
            dmin = d.min()
            dmax = d.max()

        if qac:
            out = f"{mean.value} {rms.value} {dmin.value} {dmax.value}"
            return out

        out = {"mean": mean, "median": median, "rms": rms, "min": dmin, "max": dmax}

        return out

    @log_call_to_history
    def decimate(self, n):
        """
        Decimate the `Spectrum` by n pixels.

        Parameters
        ----------
        n : int
            Decimation factor of the spectrum by returning every n-th channel.

        Returns
        -------
        s : `Spectrum`
            The decimated `Spectrum`.
        """

        if not float(n).is_integer():
            raise ValueError(f"`n` ({n}) must be an integer.")

        nchan = len(self._data)
        new_meta = deepcopy(self.meta)
        idx = np.arange(0, nchan, n)
        new_cdelt1 = self.meta["CDELT1"] * n
        cell_shift = 0.5 * (n - 1) * (self._spectral_axis.value[1] - self._spectral_axis.value[0])
        new_data = self.data[idx] * self.flux.unit
        new_meta["CDELT1"] = new_cdelt1
        new_meta["CRPIX1"] = 1.0 + (self.meta["CRPIX1"] - 1) / n + 0.5 * (n - 1) / n
        new_meta["CRVAL1"] += cell_shift

        s = Spectrum.make_spectrum(new_data, meta=new_meta, observer_location="from_meta")

        if self._baseline_model is not None:
            s._baseline_model = None

        return s

    @log_call_to_history
    def smooth(self, method="hanning", width=1, decimate=0, kernel=None):
        """
        Smooth or Convolve the `Spectrum`, optionally decimating it.

        Default smoothing is hanning.

        Parameters
        ----------
        method : string, optional
            Smoothing method. Valid are: 'hanning', 'boxcar' and
            'gaussian'. Minimum match applies.
            The default is 'hanning'.
        width : int, optional
            Effective width of the convolving kernel.  Should ideally be an
            odd number.
            For 'hanning' this should be 1, with a 0.25,0.5,0.25 kernel.
            For 'boxcar' an even value triggers an odd one with half the
            signal at the edges, and will thus not reproduce GBTIDL.
            For 'gaussian' this is the FWHM of the final frequency response.
            That is, the smoothed `Spectrum` will have an effective frequency resolution
            equal to CDELT1*`width`.
            The default is 1.
        decimate : int, optional
            Decimation factor of the spectrum by returning every decimate channel.
            -1:   no decimation
            0:    use the width parameter
            >1:   user supplied decimation (use with caution)
            The default is 0, meaning decimation is by `width`
        kernel : `~numpy.ndarray`, optional
            A numpy array which is the kernel by which the signal is convolved.
            Use with caution, as it is assumed the kernel is normalized to
            one, and is symmetric. Since width is ill-defined here, the user
            should supply an appropriate number manually.
            NOTE: not implemented yet.
            The default is None.

        Raises
        ------
        Exception
            If no valid smoothing method is given.
        ValueError
            If `width` is less than one.
            If `width` is less than the spectral resolution (in channels).
            If `decimate` is not an integer.

        Returns
        -------
        s : `Spectrum`
            The new, possibly decimated, convolved `Spectrum`.
        """
        nchan = len(self._data)
        # decimate = int(decimate) # Should we change this value and tell the user, or just error out?
        # For now, we'll error out if decimate is not an integer..

        if width < 1:
            raise ValueError(f"`width` ({width}) must be >=1.")

        # @todo  see also core.smooth() for valid_methods
        valid_methods = ["hanning", "boxcar", "gaussian"]
        this_method = minimum_string_match(method, valid_methods)
        if this_method == None:
            raise Exception(f"smooth({method}): valid methods are {valid_methods}")

        if not float(decimate).is_integer():
            raise ValueError("`decimate` must be an integer.")

        # All checks for smoothing should be completed by this point.
        # Create a new metadata dictionary to modify by smooth.
        new_meta = deepcopy(self.meta)
        md = np.ma.masked_array(self._data, self.mask)
        if this_method == "gaussian":
            if width <= self._resolution:
                raise ValueError(
                    f"`width` ({width} channels) cannot be less than the current resolution ({self._resolution} channels)."
                )
            kwidth = np.sqrt(width**2 - self._resolution**2)  # Kernel effective width.
            stddev = kwidth / 2.35482

            s1 = core.smooth(md, this_method, stddev)
        else:
            kwidth = width
            s1 = core.smooth(md, this_method, width)
        # mask = np.full(s1.shape, False)
        # in core.smooth, we fill masked values with np.nan.
        # astropy.convolve does not return a new mask, so we recreate
        # a decimated mask where values are nan
        # mask[np.where(s1 == np.nan)] = True
        # new_data = Masked(s1 * self.flux.unit, mask)
        new_data = s1 * self.flux.unit
        new_meta["FREQRES"] = np.sqrt((kwidth * self.meta["CDELT1"]) ** 2 + self.meta["FREQRES"] ** 2)

        s = Spectrum.make_spectrum(new_data, meta=new_meta, observer_location="from_meta")
        s._baseline_model = self._baseline_model

        # Now decimate if needed.
        if decimate >= 0:

            if decimate == 0:
                # Take the default decimation by `width`.
                decimate = int(abs(width))
                if not float(width).is_integer():
                    print(f"Adjusting decimation factor to be a natural number. Will decimate by {decimate}")

            s = s.decimate(decimate)

        # Update the spectral resolution in channel units.
        s._resolution = s.meta["FREQRES"] / abs(s.meta["CDELT1"])

        return s

    def shift(self, s, remove_wrap=True, fill_value=np.nan, method="fft"):
        """
        Shift the `Spectrum` by `s` channels in place.

        Parameters
        ----------
        s : float
            Number of channels to shift the `Spectrum` by.
        remove_wrap : bool
            If `False` keep channels that wrap around the edges.
            If `True` fill channels that wrap with `fill_value`.
        fill_value : float
            If `remove_wrap=True` fill channels that wrapped with this value.
        method : "fft"
            Method used to perform the fractional channel shift.
            "fft" uses a phase shift.
        """

        new_spec = self._copy()
        new_data = core.data_shift(new_spec.data, s, remove_wrap=remove_wrap, fill_value=fill_value, method=method)

        # Update data values.
        new_spec._data = new_data

        # Update metadata.
        new_spec.meta["CRPIX1"] += s

        # Update WCS.
        new_spec.wcs.wcs.crpix[0] += s

        # Update `SpectralAxis` values.
        # Radial velocity needs to be copied by hand.
        radial_velocity = deepcopy(new_spec._spectral_axis._radial_velocity)
        new_spectral_axis_values = new_spec.wcs.spectral.pixel_to_world(np.arange(new_spec.flux.shape[-1]))
        new_spec._spectral_axis = new_spec.spectral_axis.replicate(value=new_spectral_axis_values)
        new_spec._spectral_axis._radial_velocity = radial_velocity

        return new_spec

    def find_shift(self, other, units=None, frame=None):
        """
        Find the shift required to align this `Spectrum` with `other`.

        Parameters
        ----------
        other : `Spectrum`
            Target `Spectrum` to align to.
        units : {None, `astropy.units.Quantity`}
            Find the shift to align the two `Spectra` in these units.
            If `None`, the `Spectra` will be aligned using the units of
            `other`.
        frame : {None, str}
            Find the shift in this reference frame.
            If `None` will use the frame of `other`.

        Returns
        -------
        shift : float
            Number of channels that this `Spectrum` must be shifted to
            be aligned with `other`.
        """

        if not isinstance(other, Spectrum):
            raise ValueError("`other` must be a `Spectrum`.")

        if frame is not None and frame not in astropy_frame_dict.keys():
            raise ValueError(
                f"`frame` ({frame}) not recognized. Frame must be one of {', '.join(list(astropy_frame_dict.keys()))}"
            )
        else:
            frame = other._velocity_frame

        sa = self.spectral_axis.with_observer_stationary_relative_to(frame)
        tgt_sa = other.spectral_axis.with_observer_stationary_relative_to(frame)

        if units is None:
            units = tgt_sa.unit

        sa = sa.to(units)
        tgt_sa = tgt_sa.to(units)

        cdelt1 = sa[1] - sa[0]
        shift = ((sa[0] - tgt_sa[0]) / cdelt1).value

        return shift

    def align_to(self, other, units=None, frame=None, remove_wrap=True, fill_value=np.nan, method="fft"):
        """
        Align the `Spectrum` with respect to `other`.

        Parameters
        ----------
        other : `Spectrum`
            Target `Spectrum` to align to.
        units : {None, `astropy.units.Quantity`}
            Find the shift to align the two `Spectra` in these units.
            If `None`, the `Spectra` will be aligned using the units of
            `other`.
        frame : {None, str}
            Find the shift in this reference frame.
            If `None` will use the frame of `other`.
        remove_wrap : bool
            If `True` allow spectrum to wrap around the edges.
            If `False` fill channels that wrap with `fill_value`.
        fill_value : float
            If `wrap=False` fill channels that wrapped with this value.
        method : "fft"
            Method used to perform the fractional channel shift.
            "fft" uses a phase shift.
        """

        s = self.find_shift(other, units=units, frame=frame)
        return self.shift(s, remove_wrap=remove_wrap, fill_value=fill_value, method=method)

    @property
    def equivalencies(self):
        """Get the spectral axis equivalencies that can be used in converting the axis
        between km/s and frequency or wavelength"""
        equiv = u.spectral()
        sa = self.spectral_axis
        if sa.doppler_rest is not None:
            rfq = sa.doppler_rest
        elif "RESTFREQ" in self.meta:
            cunit1 = self.meta.get("CUNIT1", self.wcs.wcs.cunit[0])
            # @todo this could be done with a dict str->function
            rfq = self.meta["RESTFREQ"] * cunit1  # WCS wants no E
        else:
            rfq = None
        if rfq is not None:
            equiv.extend(get_spectral_equivalency(rfq, self.velocity_convention))
        return equiv

    @property
    def target(self):
        return self._target

    @property
    def observer(self):
        """
        Returns
        -------
            observer : `~astropy.coordinates.BaseCoordinateFrame` or derivative
            The coordinate frame of the observer if present.
        """
        return self._observer

    @property
    def velocity_frame(self):
        """String representation of the velocity frame"""
        return self._velocity_frame

    @property
    def doppler_convention(self):
        """String representation of the velocity (Doppler) convention"""
        return self.velocity_convention

    def axis_velocity(self, unit=KMS):
        """Get the spectral axis in velocity units.
        *Note*: This is not the same as `Spectrum.velocity`, which includes the source radial velocity.

        Parameters
        ----------
        unit : `~astropy.units.Quantity` or str that can be converted to Quantity
                The unit to which the axis is to be converted
        Returns
        -------
        velocity : `~astropy.units.Quantity`
                The converted spectral axis velocity
        """
        return self._spectral_axis.to(unit)

    def velocity_axis_to(self, unit=KMS, toframe=None, doppler_convention=None):
        """
        Parameters
        ----------
        unit : `~astropy.units.Quantity` or str that can be converted to Quantity
            The unit to which the axis is to be converted

        toframe : str
            The coordinate frame to convert to, e.g. 'hcrs', 'icrs'

        doppler_convention : str
            The Doppler velocity covention to use, one of 'optical', 'radio', or 'rest'

        Returns
        -------
        test_spectrum.pyvelocity : `~astropy.units.Quantity`
            The converted spectral axis velocity
        """
        if toframe is not None and toframe != self.velocity_frame:
            s = self.with_frame(toframe)
        else:
            s = self
        if doppler_convention is not None:
            return s._spectral_axis.to(unit=unit, doppler_convention=doppler_convention).to(unit)
        else:
            return s.axis_velocity(unit)

    def get_velocity_shift_to(self, toframe):
        if self._target is None:
            raise Exception("Can't calculate velocity because Spectrum.target is None")
        return get_velocity_in_frame(self._target, toframe, self._observer, self._obstime)

    @log_call_to_history
    def set_frame(self, toframe):
        """Set the sky coordinate and doppler tracking reference frame of this Spectrum. The header 'CTYPE1' will be changed accordingly.

        To make a copy of this Spectrum with new coordinate referece frmae instead, use `with_frame`.

        Parameters
        ----------
        toframe - str, or ~astropy.coordinates.BaseCoordinateFrame, or ~astropy.coordinates.SkyCoord
            The coordinate reference frame identifying string, as used by astropy, e.g. 'hcrs', 'icrs', etc.,
            or an actual coordinate system instance
        """

        tfl = toframe
        if isinstance(toframe, str):
            tfl = toframe.lower()
            tfl = astropy_convenience_frame_names.get(tfl, tfl)
            if "itrs" in tfl:
                if isinstance(self._observer, ITRS):
                    return  # nothing to be done, we already have the correct axis
                raise ValueError(
                    "For topographic or ITRS coordaintes, you must supply a full astropy Coordinate instance."
                )
            elif self._velocity_frame == tfl:
                return  # the frame is already the requested frame

        self._spectral_axis = self._spectral_axis.with_observer_stationary_relative_to(tfl)
        self._observer = self._spectral_axis.observer
        # This line is commented because:
        # SDFITS defines CTYPE1 as always being the TOPO frequency.
        # See Issue #373 on GitHub.
        # self._meta["CTYPE1"] = change_ctype(self._meta["CTYPE1"], toframe)
        if isinstance(tfl, str):
            self._velocity_frame = tfl
        else:
            self._velocity_frame = tfl.name
        # While it is incorrect to change CTYPE1, it is reasonable to change VELDEF.
        # SDFITS defines CTYPE1 as always being the TOPO frequency.
        # See Issue #373 on GitHub.
        self.meta["VELDEF"] = change_ctype(self.meta["VELDEF"], self._velocity_frame)

    def with_frame(self, toframe):
        """Return a copy of this `Spectrum` with a new coordinate reference frame.

        Parameters
        ----------
        toframe - str, `~astropy.coordinates.BaseCoordinateFrame`, or `~astropy.coordinates.SkyCoord`
            The coordinate reference frame identifying string, as used by astropy, e.g. 'hcrs', 'icrs', etc.,
            or an actual coordinate system instance

        Returns
        -------
        spectrum : `Spectrum`
            A new `Spectrum` object with the rest frame set to `toframe`.
        """

        s = self._copy()
        s.set_frame(toframe)
        return s

    @log_call_to_history
    def set_convention(self, doppler_convention):
        """Set the velocity convention of this `Spectrum`.  The spectral axis of this `Spectrum` will be replaced
        with a new spectral axis with the input velocity convention.  The header 'VELDEF' value will
        be changed accordingly.

        To make a copy of this `Spectrum` with a new velocity convention instead, use `with_velocity_convention`.

        Parameters
        ----------
        doppler_convention - str
            The velocity convention, one of 'radio', 'optical', 'relativistic'

        """
        # replicate() gives the same asnwer as
        # self._spectral_axis.to(unit=self._spectral_axis.unit, doppler_convention=doppler_convention)
        new_sp_axis = self.spectral_axis.replicate(doppler_convention=doppler_convention)
        self._spectral_axis = new_sp_axis
        self.meta["VELDEF"] = replace_convention(self.meta["VELDEF"], doppler_convention)

    def with_velocity_convention(self, doppler_convention):
        """Returns a copy of this `Spectrum` with the input velocity convention.  The header 'VELDEF' value will
        be changed accordingly.

        Parameters
        ----------
        doppler_convention - str
            The velocity convention, one of 'radio', 'optical', 'relativistic'

        Returns
        -------
        spectrum : `Spectrum`
            A new `Spectrum` object with `doppler_convention` as its Doppler convention.
        """
        if False:
            # this doesn't work.
            # for some reason, the axis velocity
            # still contains the difference between TOPO and observed frame
            s = self.__class__(
                flux=self.flux,
                wcs=self.wcs,
                meta=self.meta,
                velocity_convention=doppler_convention,
                target=self._target,
                observer=self._spectral_axis.observer,
            )
            s.meta["VELDEF"] = replace_convention(self.meta["VELDEF"], doppler_convention)
        s = self._copy(velocity_convention=doppler_convention)
        s.set_convention(doppler_convention)
        return s

    def savefig(self, file, **kwargs):
        """Save the plot"""
        if self._plotter is None:
            raise Exception("You have to invoke plot() first")
        self._plotter.figure.savefig(file, **kwargs)

    def _write_table(self, fileobj, format, **kwargs):
        """
        Write this `Spectrum` as an ~astropy.table.Table. The output columns will be
            - frequency or velocity - in  the Spectrum's spectral axis units, or converted with `xaxis_unit` keyword
            - flux - in the Spectrum's flux units, or converted with `yaxis_unit` keyword
            - uncertainty - The channel-based uncertainty or zeroes if uncertainty was not defined
            - weights - the channel weights
            - mask - 0 is unmasked ('good'), 1 is masked ('bad')
            - baseline model value - the value of the baseline model at each x axis point, regardless of whether it
              has been subtracted from the Spectrum or not. This will be all zeroes if no baseline model
              has been computed.

        Parameters
        ----------
        fileobj : str, file-like or `pathlib.Path`
            File to write to.  If a file object, must be opened in a
            writeable mode.
        format : str
            The output format. Must be a format supported by ~astropy.table.Table.write
        kwargs : variable
            Additional keyword arguments supported by ~astropy.table.Table.write

        """
        # @todo support xaxis_unit, yaxis_unit, flux_unit
        # xaxis_unit : str or :class:`astropy.units.Unit`
        # yaxis_unit : str or :class:`astropy.units.Unit`
        # flux_unit : str or :class:`astropy.units.Unit`
        if self._baseline_model is None:
            bl = np.zeros_like(self.flux)
            bldesc = "Fitted baseline value at given channel (was not defined)"
        else:
            bl = self._baseline_model(self._spectral_axis)
            bldesc = "Fitted baseline value at given channel"
        mask = self.mask * 1
        if self.uncertainty is None:
            unc = np.zeros(self.flux.shape)  # , mask=False)
            udesc = "Flux uncertainty (was not defined)"
        else:
            unc = self.uncertainty.quantity
            udesc = "Flux uncertainty"
        outarray = [self.spectral_axis, self.flux, unc, self.weights, mask, bl]
        description = ["Spectral axis", "Flux", udesc, "Channel weights", "Mask 0=unmasked, 1=masked", bldesc]
        # remove FITS reserve keywords
        meta = deepcopy(self.meta)
        meta.pop("NAXIS1", None)
        meta.pop("TDIM7", None)
        meta.pop("TUNIT7", None)
        meta["HISTORY"] = self.history  # use property not _history to ensure ascii
        meta["COMMENT"] = self.comments
        if format == "mrt":
            ulab = "e_flux"  # MRT convention that error on X is labeled e_X
        else:
            ulab = "uncertainty"
        outnames = ["spectral_axis", "flux", ulab, "weight", "mask", "baseline"]
        if "ipac" in format:
            # IPAC format wants a crazy dictionary style.
            d = {}
            for k, v in meta.items():
                d[k] = {"value": v}
            t = Table(outarray, names=outnames, meta={"keywords": d}, descriptions=description)
        else:
            t = Table(outarray, names=outnames, meta=meta, descriptions=description)
        # for now ignore complaints about keywords until we clean them up.
        # There are some that are more than 8 chars that should be fixed in GBTFISLOAD
        warnings.simplefilter("ignore", VerifyWarning)
        t.write(fileobj, format=format, **kwargs)

    def _copy(self, **kwargs):
        """
        Perform deep copy operations on each attribute of the ``Spectrum``
        object.
        This overrides the ``specutils.Spectrum1D`` method so that
        target and observer attributes get copied.
        """
        alt_kwargs = dict(
            flux=deepcopy(self.flux),
            spectral_axis=deepcopy(self.spectral_axis),
            uncertainty=deepcopy(self.uncertainty),
            wcs=deepcopy(self.wcs),
            mask=deepcopy(self.mask),
            meta=deepcopy(self.meta),
            unit=deepcopy(self.unit),
            velocity_convention=deepcopy(self.velocity_convention),
            rest_value=deepcopy(self.rest_value),
            target=deepcopy(self.target),
            observer=deepcopy(self.observer),
        )

        alt_kwargs.update(kwargs)

        s = self.__class__(**alt_kwargs)
        # s.topocentric = is_topocentric(meta["CTYPE1"])  # GBT-specific to use CTYPE1 instead of VELDEF
        return s

    @classmethod
    def fake_spectrum(cls, nchan=1024, seed=None, **kwargs):
        """
        Create a fake spectrum, useful for simple testing. A default header is
        created, which may be modified with kwargs.

        Parameters
        ----------
        nchan : int, optional
            Number of channels. The default is 1024.

        seed : {None, int, array_like[ints], `numpy.random.SeedSequence`, `numpy.random.BitGenerator`, `numpy.random.Generator`}, optional
            A seed to initialize the `BitGenerator`. If None, then fresh, unpredictable entropy will be pulled from the OS.
            If an int or array_like[ints] is passed, then all values must be non-negative and will be passed to
            `SeedSequence` to derive the initial `BitGenerator` state. One may also pass in a `SeedSequence` instance.
            Additionally, when passed a `BitGenerator`, it will be wrapped by `Generator`. If passed a `Generator`, it will
            be returned unaltered.
            The default is `None`.

        **kwargs: dict or key=value
            Metadata to put in the header.  If the key exists already in
            the default header, it will be replaced. Otherwise the key and value will be
            added to the header. Keys are case insensitive.

        Returns
        -------
        spectrum : `Spectrum`
            The spectrum object
        """

        rng = np.random.default_rng(seed)
        data = rng.random(nchan) * u.K
        meta = {
            "OBJECT": "NGC2415",
            "BANDWID": 23437500.0,
            "DATE-OBS": "2021-02-10T07:38:37.50",
            "DURATION": 0.9982445,
            "EXPOSURE": 732.1785161896237,
            "TSYS": 17.930595470605255,
            "CTYPE1": "FREQ-OBS",
            "CRVAL1": 1402544936.7749996,
            "CRPIX1": float(nchan) / 2.0,
            "CDELT1": -715.2557373046875,
            "CTYPE2": "RA",
            "CRVAL2": 114.23878994411744,
            "CTYPE3": "DEC",
            "CRVAL3": 35.24315395841497,
            "CRVAL4": -6,
            "OBSERVER": "Michael Fanelli",
            "OBSID": "unknown",
            "SCAN": 152,
            "OBSMODE": "OnOff:PSWITCHON:TPWCAL",
            "FRONTEND": "Rcvr1_2",
            "TCAL": 1.4551637172698975,
            "VELDEF": "OPTI-HEL",
            "VFRAME": 15264.39118499772,
            "RVSYS": 3775382.910954342,
            "OBSFREQ": 1402544936.7749996,
            "LST": 42101.90296246562,
            "AZIMUTH": 285.9514963267411,
            "ELEVATIO": 42.100623613548194,
            "TAMBIENT": 270.4,
            "PRESSURE": 696.2290227048372,
            "HUMIDITY": 0.949,
            "RESTFREQ": 1420405751.7,
            "FREQRES": 715.2557373046875,
            "EQUINOX": 2000.0,
            "RADESYS": "FK5",
            "TRGTLONG": 114.2375,
            "TRGTLAT": 35.24194444444444,
            "SAMPLER": "A2_0",
            "FEED": 1,
            "SRFEED": 0,
            "FEEDXOFF": 0.0,
            "FEEDEOFF": 0.0,
            "SUBREF_STATE": 1,
            "SIDEBAND": "L",
            "PROCSEQN": 1,
            "PROCSIZE": 2,
            "PROCSCAN": "ON",
            "PROCTYPE": "SIMPLE",
            "LASTON": 152,
            "LASTOFF": 0,
            "TIMESTAMP": "2021_02_10_07:38:37",
            "QD_BAD": -1,
            "QD_METHOD": "",
            "VELOCITY": 3784000.0,
            "DOPFREQ": 1420405751.7,
            "ADCSAMPF": 3000000000.0,
            "VSPDELT": 65536.0,
            "VSPRVAL": 19.203125,
            "VSPRPIX": 16385.0,
            "SIG": "T",
            "CAL": "F",
            "CALTYPE": "LOW",
            "CALPOSITION": "Unknown",
            "IFNUM": 0,
            "PLNUM": 0,
            "FDNUM": 0,
            "HDU": 1,
            "BINTABLE": 0,
            "ROW": 2,
            "DATE": "2022-02-14T17:25:01",
            "ORIGIN": "NRAO Green Bank",
            "TELESCOP": "NRAO_GBT",
            "INSTRUME": "VEGAS",
            "SDFITVER": "sdfits ver1.22",
            "FITSVER": "1.9",
            "CTYPE4": "STOKES",
            "PROJID": "TGBT21A_501_11",
            "BACKEND": "VEGAS",
            "SITELONG": -79.83983,
            "SITELAT": 38.43312,
            "SITEELEV": 824.595,
            "EXTNAME": "SINGLE DISH",
            "FITSINDEX": 0,
            "PROC": "OnOff",
            "OBSTYPE": "PSWITCHON",
            "SUBOBSMODE": "TPWCAL",
            "INTNUM": 0,
            "CUNIT1": "Hz",
            "CUNIT2": "deg",
            "CUNIT3": "deg",
            "RESTFRQ": 1420405751.7,
            "MEANTSYS": 17.16746070048293,
            "WTTSYS": 17.16574907094451,
        }
        for k, v in kwargs.items():
            meta[k.upper()] = v
        return Spectrum.make_spectrum(data, meta, observer_location=Observatory["GBT"])

    # @todo allow observer or observer_location.  And/or sort this out in the constructor.
    @classmethod
    def make_spectrum(cls, data, meta, use_wcs=True, observer_location=None, observer=None):
        # , shift_topo=False):
        """Factory method to create a `Spectrum` object from a data and header.  The the data are masked,
        the `Spectrum` mask will be set to the data mask.

        Parameters
        ----------
        data :  `~numpy.ndarray`
            The data array. See `~specutils.Spectrum1D`
        meta : dict
            The metadata, typically derived from an SDFITS header.
            Required items in `meta` are 'CTYPE[123]','CRVAL[123]', 'CUNIT[123]', 'VELOCITY', 'EQUINOX', 'RADESYS'
        use_wcs : bool
            If True, create a WCS object from `meta`
            Default: True
        observer_location : `~astropy.coordinates.EarthLocation` or str
            Location of the observatory. See `~dysh.coordinates.Observatory`.
            This will be transformed to `~astropy.coordinates.ITRS` using the time of observation DATE-OBS or MJD-OBS in `meta`.
            If this parameter is given the special str value 'from_meta', then an observer_location
            will be created from SITELONG, SITELAT, and SITEELEV in the meta dictionary.
        observer : `~astropy.coordinates.BaseCoordinateFrame`
            Coordinate frame for the observer.
            Will be ignored if DATE-OBS or MJD-OBS are present in `meta` and `observer_location` is not `None`.

        Returns
        -------
        spectrum : `Spectrum`
            The spectrum object
        """
        # @todo generic check_required method since I now have this code in two places (coordinates/core.py).
        # @todo requirement should be either DATE-OBS or MJD-OBS, but make_target() needs to be updated
        # in that case as well.
        _required = set(
            [
                "CRVAL1",
                "CRVAL2",
                "CRVAL3",
                "CTYPE1",
                "CTYPE2",
                "CTYPE3",
                "CUNIT1",
                "CUNIT2",
                "CUNIT3",
                "VELOCITY",
                "EQUINOX",
                "RADESYS",
                "DATE-OBS",
                "VELDEF",
                "RESTFRQ",
            ]
        )

        if not _required <= meta.keys():
            raise ValueError(f"Header (meta) is missing one or more required keywords: {_required}")

        # @todo WCS is expensive.
        # Possibly figure how to calculate spectral_axis instead.
        # @todo allow fix=False in WCS constructor?
        if use_wcs:
            with warnings.catch_warnings():
                # Skip warnings about DATE-OBS being converted to MJD-OBS.
                warnings.filterwarnings("ignore", category=FITSFixedWarning)
                # Skip warnings FITS keywords longer than 8 chars or containing
                # illegal characters (like _).
                warnings.filterwarnings("ignore", category=VerifyWarning)
                # lists are problematic in constructor to WCS
                # so temporarily remove history and comment values
                savehist = meta.pop("HISTORY", None)
                savecomment = meta.pop("COMMENT", None)
                if savecomment is None:
                    savecomment = meta.pop("comments", None)
                wcs = WCS(header=meta)
                if savehist is not None:
                    meta["HISTORY"] = savehist
                if savecomment is not None:
                    meta["COMMENT"] = savecomment
                # It would probably be safer to add NAXISi to meta.
                if wcs.naxis > 3:
                    wcs.array_shape = (0, 0, 0, len(data))
                # For some reason these aren't identified while creating the WCS object.
                if "SITELONG" in meta.keys():
                    wcs.wcs.obsgeo[:3] = meta["SITELONG"], meta["SITELAT"], meta["SITEELEV"]
                # Reset warnings.
        else:
            wcs = None
        # is_topo = is_topocentric(meta["CTYPE1"])  # GBT-specific to use CTYPE1 instead of VELDEF
        target = make_target(meta)
        vc = veldef_to_convention(meta["VELDEF"])

        # Define an observer as needed.
        if observer is not None:
            obsitrs = observer
        elif observer_location is not None and (meta.get("DATE-OBS") or meta.get("MJD-OBS")) is not None:
            obstime = Time(meta.get("DATE-OBS") or meta.get("MJD-OBS"))
            if observer_location == "from_meta":
                try:
                    observer_location = Observatory.get_earth_location(
                        meta["SITELONG"], meta["SITELAT"], meta["SITEELEV"]
                    )
                except KeyError as ke:
                    raise Exception(f"Not enough info to create observer_location: {ke}")
            obsitrs = SpectralCoord._validate_coordinate(
                attach_zero_velocities(observer_location.get_itrs(obstime=obstime))
            )
        else:
            warnings.warn(
                "'meta' does not contain DATE-OBS or MJD-OBS. Spectrum won't be convertible to certain coordinate"
                " frames"
            )
            obsitrs = None

        if np.ma.is_masked(data):
            s = cls(
                flux=data,
                wcs=wcs,
                meta=meta,
                velocity_convention=vc,
                radial_velocity=target.radial_velocity,
                rest_value=meta["RESTFRQ"] * u.Hz,
                observer=obsitrs,
                target=target,
                mask=data.mask,
            )
        else:
            s = cls(
                flux=data,
                wcs=wcs,
                meta=meta,
                velocity_convention=vc,
                radial_velocity=target.radial_velocity,
                rest_value=meta["RESTFRQ"] * u.Hz,
                observer=obsitrs,
                target=target,
            )
        # For some reason, Spectrum1D.spectral_axis created with WCS do not inherit
        # the radial velocity. In fact, they get no radial_velocity attribute at all!
        # This method creates a new spectral_axis with the given radial velocity.
        if observer_location is None and observer is None:
            s.set_radial_velocity_to(target.radial_velocity)  # open
        return s

    def _arithmetic_apply(self, other, op, handle_meta, **kwargs):
        if isinstance(other, NDCube):
            result = op(other, **{"handle_meta": handle_meta})
        else:
            result = op(other, **{"handle_meta": handle_meta, "meta_other_meta": False})
        self._copy_attributes(result)
        return result

    def _copy_attributes(self, other):
        """
        Copy `Spectrum` attributes after
        an arithmetic operation.
        Only copy attributes that are not modified by the arithmetic.
        I.e., do not copy the "_flux" attribute.
        """
        for k, v in vars(self).items():
            if k not in IGNORE_ON_COPY:
                vars(other)[k] = deepcopy(v)

    def __add__(self, other):
        op = self.add
        handle_meta = self._add_meta
        if not isinstance(other, (NDCube, u.Quantity)):
            try:
                other = u.Quantity(other, unit=self.unit)
            except TypeError:
                return NotImplemented
        result = self._arithmetic_apply(other, op, handle_meta)
        return result

    __radd__ = __add__

    def __sub__(self, other):
        op = self.subtract
        handle_meta = self._add_meta
        if not isinstance(other, (NDCube, u.Quantity)):
            try:
                other = u.Quantity(other, unit=self.unit)
            except TypeError:
                return NotImplemented
        result = self._arithmetic_apply(other, op, handle_meta)
        return result

    def __rsub__(self, other):
        return -1 * (self - other)

    def __mul__(self, other):
        op = self.multiply
        handle_meta = self._mul_meta
        if not isinstance(other, NDCube):
            other = u.Quantity(other)
        result = self._arithmetic_apply(other, op, handle_meta)
        return result

    __rmul__ = __mul__

    def __div__(self, other):
        op = self.divide
        handle_meta = self._div_meta
        if not isinstance(other, NDCube):
            other = u.Quantity(other)
        result = self._arithmetic_apply(other, op, handle_meta)
        return result

    def __truediv__(self, other):
        op = self.divide
        handle_meta = self._div_meta
        if not isinstance(other, NDCube):
            other = u.Quantity(other)
        result = self._arithmetic_apply(other, op, handle_meta)
        return result

    def _add_meta(self, operand, operand2, **kwargs):
        kwargs.setdefault("other_meta", True)
        meta = deepcopy(operand)
        if kwargs["other_meta"]:
            meta["EXPOSURE"] = operand["EXPOSURE"] + operand2["EXPOSURE"]
            meta["DURATION"] = operand["DURATION"] + operand2["DURATION"]

        return meta

    def _mul_meta(self, operand, operand2=None, **kwargs):
        # TBD
        return deepcopy(operand)

    def _div_meta(self, operand, operand2=None, **kwargs):
        # TBD
        return deepcopy(operand)

    def __getitem__(self, item):

        def q2idx(q, wcs, spectral_axis, coo, sto):
            """Quantity to index."""
            if "velocity" in u.get_physical_type(q):
                # Convert from velocity to channel.
                idx = vel2idx(q, wcs, spectral_axis, coo, sto)
            elif "length" in u.get_physical_type(q):
                # Convert wavelength to channel.
                idx = wav2idx(q, wcs, spectral_axis, coo, sto)
            elif "frequency" in u.get_physical_type(q):
                # Convert from frequency to channel.
                idxs = wcs.world_to_pixel(coo, q, sto)
                idx = int(np.round(idxs[0]))
            return idx

        def vel2idx(vel, wcs, spectral_axis, coo, sto):
            eq = get_spectral_equivalency(spectral_axis.doppler_rest, spectral_axis.doppler_convention)
            # Make `vel` a `SpectralCoord`.
            vel_sp = spectral_axis.to(unit=vel.unit).replicate(value=vel.value, unit=vel.unit)
            with u.set_enabled_equivalencies(eq):
                idxs = wcs.world_to_pixel(coo, vel_sp, sto)
            return int(np.round(idxs[0]))

        def wav2idx(wav, wcs, spectral_axis, coo, sto):
            with u.set_enabled_equivalencies(u.spectral()):
                wav_sp = spectral_axis.to(unit=wav.unit).replicate(value=wav.value, unit=wav.unit)
                idxs = wcs.world_to_pixel(coo, wav_sp, sto)
            return int(np.round(idxs[0]))

        # Only slicing along the spectral coordinate.
        # Assumes that the WCS is in frequency.
        if isinstance(item, slice):
            if item.step is not None and item.step != 1:
                raise NotImplementedError(
                    "Cannot slice with a step other than 1. Try smoothing and decimating if you want to downsample."
                )

            spectral_axis = self.spectral_axis
            # Use the WCS to convert from world to pixel values.
            wcs = self.wcs
            # We need a sky location to convert incorporating velocity shifts.
            coo = SkyCoord(
                wcs.wcs.crval[wcs.wcs.lng] * wcs.wcs.cunit[wcs.wcs.lng],
                wcs.wcs.crval[wcs.wcs.lat] * wcs.wcs.cunit[wcs.wcs.lat],
                frame="fk5",
            )
            # Same for the Stokes axis.
            sto = StokesCoord(0)
            start = item.start
            stop = item.stop

            # Start.
            if isinstance(start, u.Quantity):
                start_idx = q2idx(start, wcs, spectral_axis, coo, sto)
            else:
                start_idx = start

            # Stop.
            if isinstance(stop, u.Quantity):
                stop_idx = q2idx(stop, wcs, spectral_axis, coo, sto)
            else:
                stop_idx = stop

        else:
            raise NotImplementedError("Use a slice for slicing.")

        # Slicing uses NumPY ordering by default.
        sliced_wcs = wcs[0:1, 0:1, 0:1, start_idx:stop_idx]

        # Update meta.
        meta = self.meta.copy()
        head = sliced_wcs.to_header()
        for k in ["CRPIX1", "CRVAL1"]:
            meta[k] = head[k]

        # New Spectrum.
        return self.make_spectrum(
            Masked(self.flux[start_idx:stop_idx], self.mask[start_idx:stop_idx]),
            meta=meta,
            observer_location=Observatory[meta["TELESCOP"]],
        )

    def average(self, spectra, weights="tsys", align=False):
        r"""
        Average this `Spectrum` with `spectra`.
        The resulting `average` will have an exposure equal to the sum of the exposures,
        and coordinates and system temperature equal to the weighted average of the coordinates and system temperatures.

        Parameters
        ----------
        spectra : list of `Spectrum`
            Spectra to be averaged. They must have the same number of channels.
            No checks are done to ensure they are aligned.
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'
        align : bool
            If `True` align the `spectra` to itself.
            This uses `Spectrum.align_to`.

        Returns
        -------
        average : `Spectrum`
            Averaged spectra.
        """

        if type(spectra) is not list:
            spectra = [spectra]

        spectra += [self]

        return average_spectra(spectra, weights=weights, align=align)


# @todo figure how how to document write()
####################################################################
# There is probably a less brute-force way to do this but I haven't
# been able to figure it out.  astropy.io.registry tools are not
# well explained.  register_writer 'consumes' the format keyword, so
# it cannot be passed along via a single overaarching write method,
# e.g., spectrum_writer()
####################################################################
def ascii_spectrum_writer_basic(spectrum, fileobj, **kwargs):
    spectrum._write_table(fileobj, format="ascii.basic", **kwargs)


def ascii_spectrum_writer_commented_header(spectrum, fileobj, **kwargs):
    spectrum._write_table(fileobj, format="ascii.commented_header", **kwargs)


def ascii_spectrum_writer_fixed_width(spectrum, fileobj, **kwargs):
    spectrum._write_table(fileobj, format="ascii.fixed_width", **kwargs)


def ascii_spectrum_writer_ipac(spectrum, fileobj, **kwargs):
    spectrum._write_table(fileobj, format="ascii.ipac", **kwargs)


def spectrum_writer_votable(spectrum, fileobj, **kwargs):
    spectrum._write_table(fileobj, format="votable", **kwargs)


def spectrum_writer_ecsv(spectrum, fileobj, **kwargs):
    spectrum._write_table(fileobj, format="ascii.ecsv", **kwargs)


def spectrum_writer_mrt(spectrum, fileobj, **kwargs):
    spectrum._write_table(fileobj, format="mrt", **kwargs)


def spectrum_writer_fits(spectrum, fileobj, **kwargs):
    spectrum._write_table(fileobj, format="fits", **kwargs)


def _read_table(fileobj, format, **kwargs):
    # GBTIDL format example:
    #    Scan:    152          NGC2415 2021-02-10 +07 38 37.5
    #                             Ta
    #     GHz-HEL                 YY
    #   1.4143356980054205     -0.1042543
    #   1.4143349827132639      0.0525000
    #
    # Note we don't get the full coordinates, just RA!
    if format == "gbtidl":
        # Read it into a pandas dataframe, which will have one column
        # with a name like 'Scan:    152          NGC2415 2021-02-10 +07 38 37.5'
        # We can split out the data array
        # We do not use e.g., Table.read(fileobs,data_start=3) because we need to
        # get the header info and don't want to open the file twice.
        # Note that functions like Table.read, pandas.read*, np.load_txt also natively
        # recognize compressed files, which is why we aren't using raw Python I/O
        # which does not.
        # t = Table.read(fileobj, format="ascii.fixed_width")  # ,data_start=3,header_start=3)
        df = pd.read_table(fileobj)
        df2 = df[df.columns[0]][2:].str.split(expand=True).astype(float).add_prefix("col")
        spectral_axis = df2["col0"]
        flux = df2["col1"]
        # Parse the first line of the header and put into meta
        tmp, scan, target, date, ra = df.columns[0].split(maxsplit=4)
        meta = {}
        meta["SCAN"] = int(scan)
        meta["OBJECT"] = target
        meta["DATE-OBS"] = Time(date).to_string()
        c = SkyCoord(ra + " 0 0 0.0", unit=(u.hour, u.deg))
        meta["RA"] = c.ra.degree
        # Parse the 2nd link of the header to get the veldef and flux unit
        # Sometimes velocity_convention isn't there!
        h2 = df[df.columns[0]][0].split()
        if len(h2) == 2:
            velocity_convention = h2[0][0:4]
            flux_unit = h2[1]
        else:
            velocity_convention = "None"
            flux_unit = h2[0]
        if flux_unit == "Ta":
            fu = u.K
        elif flux_unit == "Counts":
            fu = u.ct
        else:
            fu = u.dimensionless_unscaled
        # parse the 3rd line column names, We only care about the ferquency axis
        h3 = df[df.columns[0]][1].split()
        units, vd = h3[0].split("-")
        spectral_axis = spectral_axis.values * u.Unit(units)
        meta["VELDEF"] = velocity_convention + "-" + vd
        meta["POL"] = h3[1]

        s = Spectrum(flux=flux.values * fu, spectral_axis=spectral_axis, meta=meta)
        return s

    t = Table.read(fileobj, format=format, **kwargs)
    f = t["flux"].value * t["flux"].unit
    return Spectrum.make_spectrum(f, meta=t.meta, observer_location="from_meta")


def spectrum_reader_ecsv(fileobj, **kwargs):
    return _read_table(fileobj, format="ascii.ecsv", **kwargs)


def spectrum_reader_fits(fileobj, **kwargs):
    return _read_table(fileobj, format="fits", **kwargs)


def spectrum_reader_gbtidl(fileobj, **kwargs):
    return _read_table(fileobj, format="gbtidl", **kwargs)


with registry.delay_doc_updates(Spectrum):
    # WRITERS
    registry.register_writer("ascii.basic", Spectrum, ascii_spectrum_writer_basic)
    registry.register_writer("basic", Spectrum, ascii_spectrum_writer_basic)
    registry.register_writer("ascii.commented_header", Spectrum, ascii_spectrum_writer_commented_header)
    registry.register_writer("commented_header", Spectrum, ascii_spectrum_writer_commented_header)
    registry.register_writer("ascii.fixed_width", Spectrum, ascii_spectrum_writer_fixed_width)
    registry.register_writer("fixed_width", Spectrum, ascii_spectrum_writer_fixed_width)
    registry.register_writer("ascii.ipac", Spectrum, ascii_spectrum_writer_ipac)
    registry.register_writer("ipac", Spectrum, ascii_spectrum_writer_ipac)
    registry.register_writer("votable", Spectrum, spectrum_writer_votable)
    registry.register_writer("ecsv", Spectrum, spectrum_writer_ecsv)
    # UnifiedOutputRegistry.write uses all caps if format not specified
    # and it believes the desired output is ecsv
    registry.register_writer("ECSV", Spectrum, spectrum_writer_ecsv)
    registry.register_writer("mrt", Spectrum, spectrum_writer_mrt)
    registry.register_writer("fits", Spectrum, spectrum_writer_fits)
    # UnifiedOutputRegistry.write uses retrurns tabular-fits if format
    # not specified and it believes the desired output is fits.
    registry.register_writer("tabular-fits", Spectrum, spectrum_writer_fits)

    # READERS
    registry.register_reader("fits", Spectrum, spectrum_reader_fits)
    registry.register_reader("gbtidl", Spectrum, spectrum_reader_gbtidl)
    registry.register_reader("ascii.ecsv", Spectrum, spectrum_reader_ecsv)
    registry.register_reader("ecsv", Spectrum, spectrum_reader_ecsv)

    # We aren't going to support these since they don't have easily digestible metadata
    # if they have metadata at all.
    # registry.register_writer("ascii.basic", Spectrum, ascii_spectrum_reader_basic)
    # registry.register_writer("basic", Spectrum, ascii_spectrum_reader_basic)
    # registry.register_writer("ascii.commented_header", Spectrum, ascii_spectrum_reader_commented_header)
    # registry.register_writer("commented_header", Spectrum, ascii_spectrum_reader_commented_header)
    # registry.register_writer("ascii.fixed_width", Spectrum, ascii_spectrum_reader_fixed_width)
    # registry.register_writer("fixed_width", Spectrum, ascii_spectrum_reader_fixed_width)
    # registry.register_writer("ascii.ipac", Spectrum, ascii_spectrum_reader_ipac)
    # registry.register_writer("ipac", Spectrum, ascii_spectrum_reader_ipac)
    # registry.register_writer("votable", Spectrum, spectrum_reader_votable)
    # registry.register_writer("mrt", Spectrum, spectrum_reader_mrt)


def average_spectra(spectra, weights="tsys", align=False):
    r"""
    Average `spectra`. The resulting `average` will have an exposure equal to the sum of the exposures,
    and coordinates and system temperature equal to the weighted average of the coordinates and system temperatures.

    Parameters
    ----------
    spectra : list of `Spectrum`
        Spectra to be averaged. They must have the same number of channels.
        No checks are done to ensure they are aligned.
    weights: str
        'tsys' or None.  If 'tsys' the weight will be calculated as:

         :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

        Default: 'tsys'
    align : bool
        If `True` align the `spectra` to the first element.
        This uses `Spectrum.align_to`.

    Returns
    -------
    average : `Spectrum`
        Averaged spectra.
    """

    nspec = len(spectra)
    nchan = len(spectra[0].data)
    shape = (nspec, nchan)
    data_array = np.ma.empty(shape, dtype=float)
    wts = np.empty(shape, dtype=float)
    exposures = np.empty(nspec, dtype=float)
    tsyss = np.empty(nspec, dtype=float)
    xcoos = np.empty(nspec, dtype=float)
    ycoos = np.empty(nspec, dtype=float)
    observer = spectra[0].observer
    units = spectra[0].flux.unit

    for i, s in enumerate(spectra):
        if not isinstance(s, Spectrum):
            raise ValueError(f"Element {i} of `spectra` is not a `Spectrum`. {type(s)}")
        if units != s.flux.unit:
            raise ValueError(
                f"Element {i} of `spectra` has units {s.flux.unit}, but the first element has units {units}."
            )
        if align:
            if i > 0:
                s = s.align_to(spectra[0])
        data_array[i] = s.data
        data_array[i].mask = s.mask

        if weights == "tsys":
            wts[i] = core.tsys_weight(s.meta["EXPOSURE"], s.meta["CDELT1"], s.meta["TSYS"])
        else:
            wts[i] = 1.0
        exposures[i] = s.meta["EXPOSURE"]
        tsyss[i] = s.meta["TSYS"]
        xcoos[i] = s.meta["CRVAL2"]
        ycoos[i] = s.meta["CRVAL3"]

    data_array = np.ma.MaskedArray(data_array, mask=np.isnan(data_array) | data_array.mask, fill_value=np.nan)
    data = np.ma.average(data_array, axis=0, weights=wts)
    tsys = np.ma.average(tsyss, axis=0, weights=wts[:, 0])
    xcoo = np.ma.average(xcoos, axis=0, weights=wts[:, 0])
    ycoo = np.ma.average(ycoos, axis=0, weights=wts[:, 0])
    exposure = exposures.sum(axis=0)

    new_meta = deepcopy(spectra[0].meta)
    new_meta["TSYS"] = tsys
    new_meta["EXPOSURE"] = exposure
    new_meta["CRVAL2"] = xcoo
    new_meta["CRVAL3"] = ycoo

    averaged = Spectrum.make_spectrum(Masked(data * units, data.mask), meta=new_meta, observer=observer)

    return averaged



""" """

import pytest

from dysh.fits import gb20mfitsload
from dysh.util import get_project_testdata


class TestGB20mFITSload:

    def setup_method(self):
        self.data_path = get_project_testdata() / "20m/Skynet_60476_DR21_118886_68343.cyb.fits"
        with pytest.warns(UserWarning):
            self.sdfits = gb20mfitsload.GB20MFITSLoad(self.data_path)

    def test_find_cal(self):
        """
        Test that the indices for finding the calibration data
        are correct.
        """
        self.sdfits.select(ifnum=0, plnum=0)
        self.sdfits._find_cal_idx()
        assert self.sdfits.cal_beg_idx == 10
        assert self.sdfits.cal_end_idx == 128

    def test_obsmode(self):
        """
        Test that the 'OBSMODE' column has been updated.
        """
        assert set(self.sdfits._index["OBSMODE"]) == set(("onoff:on", "onoff:off"))

    def test_getps(self):
        """
        Test that we can use `getps`.
        """
        spec = self.sdfits.getps(ifnum=0, plnum=0)



"""
The classes that define various types of Scan and their calibration methods.
"""

import warnings
from collections import UserList
from copy import deepcopy

import astropy.units as u
import numpy as np
from astropy import constants as ac
from astropy.io.fits import BinTableHDU, Column
from astropy.table import Table, vstack
from astropy.utils.masked import Masked

from dysh.spectra import core

from ..coordinates import Observatory
from ..log import HistoricalBase, log_call_to_history, logger
from ..util import uniq
from .core import (  # fft_shift,
    average,
    find_non_blanks,
    find_nonblank_ints,
    mean_tsys,
    sq_weighted_avg,
    tsys_weight,
)
from .spectrum import Spectrum, average_spectra


class SpectralAverageMixin:
    @log_call_to_history
    def timeaverage(self, weights=None):
        r"""Compute the time-averaged spectrum for this scan.

        Parameters
        ----------
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'
        Returns
        -------
        spectrum : :class:`~spectra.spectrum.Spectrum`
            The time-averaged spectrum

        .. note::
           Data that are masked will have values set to zero.  This is a feature of `numpy.ma.average`. Data mask fill value is NaN (np.nan)
        """
        pass

    @log_call_to_history
    def polaverage(self, weights=None):
        """Average all polarizations in this Scan"""
        pass

    @log_call_to_history
    def finalspectrum(self, weights=None):
        """Average all times and polarizations in this Scan"""
        pass

    @property
    def exposure(self):
        """The array of exposure (integration) times. How the exposure is calculated
        varies for different derrived classes.

        Returns
        -------
        exposure : ~numpy.ndarray
            The exposure time in units of the EXPOSURE keyword in the SDFITS header
        """
        return self._exposure

    @property
    def delta_freq(self):
        """The array of channel frequency width"""
        return self._delta_freq

    @property
    def tsys(self):
        """The system temperature array.

        Returns
        -------
        tsys : `~numpy.ndarray`
            System temperature values in K
        """
        return self._tsys

    @property
    def tsys_weight(self):
        r"""The system temperature weighting array computed from current
        :math`T_{sys}`, :math:`t_{int}`, and :math:`\delta\nu`. See :meth:`tsys_weight`
        """
        return tsys_weight(self.exposure, self.delta_freq, self.tsys)


class ScanBase(HistoricalBase, SpectralAverageMixin):
    """This class describes the common interface to all Scan classes.
    A Scan represents one scan number, one IF, one feed, and one or more polarizations.
    Derived classes *must* implement :meth:`calibrate`.
    """

    def __init__(self, sdfits):
        HistoricalBase.__init__(self)
        self._ifnum = -1
        self._fdnum = -1
        self._nchan = -1
        self._npol = -1
        self._scan = -1
        self._pols = -1
        self._sdfits = sdfits
        # self._meta = {}

    def _validate_defaults(self):
        _required = {
            "IFNUM": self._ifnum,
            "FDNUM": self._fdnum,
            "NCHAN": self._nchan,
            "NPOL": self._npol,
            "POLS": self._pols,
            "SCAN": self._scan,
        }
        unset = []
        for k, v in _required.items():
            if v == -1:
                unset.append(k)
        if len(unset) > 0:
            raise Exception(
                f"The following required Scan attributes were not set by the derived class {self.__class__.__name__}:"
                f" {unset}"
            )
        if type(self._scan) != int:
            raise (f"{self.__class__.__name__}._scan is not an int: {type(self._scan)}")

    @property
    def scan(self):
        """
        The scan number

        Returns
        -------
        int
            The scan number of the integrations in the Scan object
        """
        return self._scan

    @property
    def nchan(self):
        """
        The number of channels in this scan

        Returns
        -------
        int
            The number of channels in this scan

        """
        return self._nchan

    @property
    def nrows(self):
        """The number of rows in this Scan

        Returns
        -------
        int
            The number of rows in this Scan
        """
        return self._nrows

    @property
    def npol(self):
        """
        The number of polarizations in this Scan

        Returns
        -------
        int
            The number of polarizations in this Scan

        """
        return self._npol

    @property
    def ifnum(self):
        """The IF number

        Returns
        -------
        int
            The index of the Intermediate Frequency
        """
        return self._ifnum

    @property
    def fdnum(self):
        """The feed number

        Returns
        -------
        int
            The index of the Feed
        """
        return self._fdnum

    @property
    def pols(self):
        """The polarization number(s)

        Returns
        -------
        list
            The list of integer polarization number(s)
        """
        return self._pols

    @property
    def is_calibrated(self):
        """
        Have the data been calibrated?

        Returns
        -------
        bool
            True if the data have been calibrated, False if not.

        """
        return self._calibrated is not None

    @property
    def meta(self):
        """
        The metadata of this Scan. The metadata is a list of dictionaries, the length of which is
        equal to the number of calibrated integrations in the Scan.

        Returns
        -------
        dict
            Dictionary containing the metadata of this Scan

        """
        return self._meta

    def _meta_as_table(self):
        """get the metadata as an astropy Table"""
        # remember, self._meta is a list of dicts,
        # and despite its name it is not Table metadata, it
        # it Table Columns values
        d = {}
        if len(self.history) != 0:
            d["HISTORY"] = self.history
        if len(self.comments) != 0:
            d["COMMENT"] = self.comments
        return Table(self._meta, meta=d)

    def _make_meta(self, rowindices):
        """
        Create the metadata for a Scan.  The metadata is a list of dictionaries, the length of which is
        equal to the number of calibrated integrations in the Scan.

        Parameters
        ----------
        rowindices : list of int
            The list of indices into the parent SDFITS index (DataFrame). These typically point to the
            indices for only the, e.g. the SIG data, since the resultant metadata will be used to make the calibrated spectra,
            which SIGREF data result in N/2 spectra.

        Returns
        -------
        None

        """
        # FITS can't handle NaN in a header, so just drop any column where NaN appears
        # Ideally this should be done on the individual record level in say, Spectrum.make_spectrum.
        df = self._sdfits.index(bintable=self._bintable_index).iloc[rowindices].dropna(axis=1, how="any")
        self._meta = df.to_dict("records")  # returns dict(s) with key = row number.
        for i in range(len(self._meta)):
            if "CUNIT1" not in self._meta[i]:
                self._meta[i][
                    "CUNIT1"
                ] = "Hz"  # @todo this is in gbtfits.hdu[0].header['TUNIT11'] but is it always TUNIT11?
            self._meta[i]["CUNIT2"] = "deg"  # is this always true?
            self._meta[i]["CUNIT3"] = "deg"  # is this always true?
            restfrq = self._meta[i]["RESTFREQ"]
            rfq = restfrq * u.Unit(self._meta[i]["CUNIT1"])
            restfreq = rfq.to("Hz").value
            self._meta[i]["RESTFRQ"] = restfreq  # WCS wants no E

    def _add_calibration_meta(self):
        """Add metadata that are computed after calibration."""
        if not self.is_calibrated:
            raise Exception("Data have to be calibrated first to add calibration metadata")
        for i in range(len(self._meta)):
            self._meta[i]["TSYS"] = self._tsys[i]
            self._meta[i]["EXPOSURE"] = self._exposure[i]
            self._meta[i]["NAXIS1"] = len(self._calibrated[i])
            self._meta[i]["TSYS"] = self._tsys[i]
            self._meta[i]["EXPOSURE"] = self.exposure[i]

    def _set_if_fd(self, df):
        """Set the IF and FD numbers from the input dataframe and
        raise an error of there are more than one

        Parameters
        ----------
        df : ~pandas.DataFrame
            The DataFrame describing the selected data
        """
        self._ifnum = uniq(df["IFNUM"])
        self._fdnum = uniq(df["FDNUM"])
        if len(self._ifnum) > 1:
            raise Exception(f"Only one IFNUM is allowed per Scan, found {self.ifnum}")
        if len(self._fdnum) > 1:
            raise Exception(f"Only one FDNUM is allowed per Scan, found {self.fdnum}")
        self._ifnum = self._ifnum[0]
        self._fdnum = self._fdnum[0]

    def calibrate(self, **kwargs):
        """Calibrate the Scan data"""
        pass

    def make_bintable(self):
        """
        Create a :class:`~astropy.io.fits.BinaryTableHDU` from the calibrated data of this Scan.


        Returns
        -------
        b : :class:`~astropy.io.fits.BinaryTableHDU`
           A FITS binary table HDU, suitable for writing out or appending to a `~astropy.io.fits.HDUList`.

        """
        # Creating a Table from the metadata dictionary and instantiating
        # the BinTableHDU with that takes care of all the tricky
        # numpy/pandas dtypes to FITS single character format string. No need
        # to figure it out oneself (which I spent far too much time trying to do before I
        # discovered this!)
        if self._calibrated is None:
            raise Exception("Data must be calibrated before writing.")
        # Table metadata aren't preserved in BinTableHDU, so we
        # have to grab them here and add them
        # data_table = self._meta_as_table()
        # table_meta = data_table.meta
        cd = BinTableHDU(data=self._meta_as_table(), name="SINGLE DISH").columns
        form = f"{np.shape(self._calibrated)[1]}E"
        cd.add_col(Column(name="DATA", format=form, array=self._calibrated))
        b = BinTableHDU.from_columns(cd, name="SINGLE DISH")
        return b

    def write(self, fileobj, output_verify="exception", overwrite=False, checksum=False):
        """
        Write an SDFITS format file (FITS binary table HDU) of the calibrated data in this Scan

        Parameters
        ----------
        fileobj : str, file-like or `pathlib.Path`
            File to write to.  If a file object, must be opened in a
            writeable mode.
        multifile: bool, optional
            If True, write to multiple files if and only if there are multiple SDFITS files in this GBTFITSLoad.
            Otherwise, write to a single SDFITS file.
        output_verify : str
            Output verification option.  Must be one of ``"fix"``,
            ``"silentfix"``, ``"ignore"``, ``"warn"``, or
            ``"exception"``.  May also be any combination of ``"fix"`` or
            ``"silentfix"`` with ``"+ignore"``, ``+warn``, or ``+exception"
            (e.g. ``"fix+warn"``).  See https://docs.astropy.org/en/latest/io/fits/api/verification.html for more info
        overwrite : bool, optional
            If ``True``, overwrite the output file if it exists. Raises an
            ``OSError`` if ``False`` and the output file exists. Default is
            ``False``.
        checksum : bool
            When `True` adds both ``DATASUM`` and ``CHECKSUM`` cards
            to the headers of all HDU's written to the file.

        Returns
        -------
        None.

        """
        self.make_bintable().writeto(name=fileobj, output_verify=output_verify, overwrite=overwrite, checksum=checksum)

    def __len__(self):
        return self._nrows


class ScanBlock(UserList, HistoricalBase, SpectralAverageMixin):
    @log_call_to_history
    def __init__(self, *args):
        UserList.__init__(self, *args)
        HistoricalBase.__init__(self)
        self._nrows = 0
        self._npol = 0
        self._timeaveraged = []
        self._polaveraged = []
        self._finalspectrum = []

    @log_call_to_history
    def calibrate(self, **kwargs):
        """Calibrate all scans in this ScanBlock"""
        for scan in self.data:
            scan.calibrate(**kwargs)

    @log_call_to_history
    def timeaverage(self, weights="tsys"):
        r"""Compute the time-averaged spectrum for all scans in this ScanBlock.

        Parameters
        ----------
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'

        Returns
        -------
        timeaverage: list of `~spectra.spectrum.Spectrum`
            List of all the time-averaged spectra

        .. note::
           Data that are masked will have values set to zero.  This is a feature of `numpy.ma.average`. Data mask fill value is NaN (np.nan)
        """
        # warnings.simplefilter("ignore", NoVelocityWarning)
        # average of the averages
        self._timeaveraged = []
        i = 0
        for scan in self.data:
            self._timeaveraged.append(scan.timeaverage(weights))
        s = average_spectra(self._timeaveraged, weights=weights)
        s.merge_commentary(self)
        return s

    @log_call_to_history
    def polaverage(self, weights="tsys"):
        # @todo rewrite this to return a spectrum as timeaverage does now.
        r"""Average all polarizations in all scans in this ScanBlock

        Parameters
        ----------
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'
        Returns
        -------
        polaverage: list of `~spectra.spectrum.Spectrum`
            List of all the polarization-averaged spectra
        """
        self._polaveraged = []
        for scan in self.data:
            self._polaveraged.append(scan.polaverage(weights))
        return self._polaveraged

    @log_call_to_history
    def finalspectrum(self, weights="tsys"):
        r"""Average all times and polarizations in all scans this ScanBlock

        Parameters
        ----------
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'
        Returns
        -------
        finalspectra: list of `~spectra.spectrum.Spectrum`
            List of all the time- and polarization-averaged spectra
        """
        self._finalspectrum = []
        for scan in self.data:
            self._finalspectrum.append(scan.finalspectrum(weights))
        return self._finalspectrum

    def write(self, fileobj, output_verify="exception", overwrite=False, checksum=False):
        """
        Write an SDFITS format file (FITS binary table HDU) of the calibrated data in this Scan

        Parameters
        ----------
        fileobj : str, file-like or `pathlib.Path`
            File to write to.  If a file object, must be opened in a
            writeable mode.
        multifile: bool, optional
            If True, write to multiple files if and only if there are multiple SDFITS files in this GBTFITSLoad.
            Otherwise, write to a single SDFITS file.
        output_verify : str
            Output verification option.  Must be one of ``"fix"``,
            ``"silentfix"``, ``"ignore"``, ``"warn"``, or
            ``"exception"``.  May also be any combination of ``"fix"`` or
            ``"silentfix"`` with ``"+ignore"``, ``+warn``, or ``+exception"
            (e.g. ``"fix+warn"``).  See https://docs.astropy.org/en/latest/io/fits/api/verification.html for more info
        overwrite : bool, optional
            If ``True``, overwrite the output file if it exists. Raises an
            ``OSError`` if ``False`` and the output file exists. Default is
            ``False``.
        checksum : bool
            When `True` adds both ``DATASUM`` and ``CHECKSUM`` cards
            to the headers of all HDU's written to the file.

        Returns
        -------
        None.

        """
        s0 = self.data[0]
        # If there is only one scan, delegate to its write method
        # this does not preserve scanblock history
        # if len(self.data) == 1:
        #    print("only one scan")
        #   s0.write(fileobj, output_verify, overwrite, checksum)
        #   return
        # Meta are the keys of the first scan's bintable except for DATA.
        # We can use this to compare with the subsequent Scan
        # keywords without have to create their bintables first.
        defaultkeys = set(s0._meta[0].keys())
        datashape = np.shape(s0._calibrated)
        tablelist = [s0._meta_as_table()]
        for scan in self.data:  # [1:]:
            # check data shapes are the same
            thisshape = np.shape(scan._calibrated)
            if thisshape != datashape:
                # @todo Variable length arrays? https://docs.astropy.org/en/stable/io/fits/usage/unfamiliar.html#variable-length-array-tables
                # or write to separate bintables.
                raise Exception(
                    f"Data shapes of scans are not equal {thisshape}!={datashape}. Can't combine Scans into single"
                    " BinTableHDU"
                )
            # check that the header keywords are the same
            diff = set(scan._meta[0].keys()) - defaultkeys
            if len(diff) > 0:
                raise Exception(
                    f"Scan header keywords are not the same. These keywords were not present in all Scans: {diff}."
                    " Can't combine Scans into single BinTableHDU"
                )
            tablelist.append(scan._meta_as_table())
        # now do the same trick as in Scan.write() of adding "DATA" to the coldefs
        # astropy Tables can be concatenated with vstack thankfully.
        table = vstack(tablelist, join_type="exact")
        # need to preserve table.meta because it gets lost in created of "cd" ColDefs
        table_meta = table.meta
        cd = BinTableHDU(table, name="SINGLE DISH").columns
        data = np.concatenate([c._calibrated for c in self.data])
        form = f"{np.shape(data)[1]}E"
        cd.add_col(Column(name="DATA", format=form, array=data))
        b = BinTableHDU.from_columns(cd, name="SINGLE DISH")
        # preserve any meta
        for k, v in table_meta.items():
            if k == "HISTORY" or k == "COMMENT":
                continue  # will deal with these later
            b.header[k] = v
        for h in self._history:
            b.header["HISTORY"] = h
        for c in self._comments:
            b.header["COMMENT"] = c

        b.writeto(name=fileobj, output_verify=output_verify, overwrite=overwrite, checksum=checksum)


class TPScan(ScanBase):
    """GBT specific version of Total Power Scan

    Parameters
    ----------
    gbtfits : `~fits.sdfitsload.SDFITSLoad`
        input SDFITSLoad object
    scan: int
        scan number
    sigstate : bool
        Select the signal state used to form the data.  True means select sig='T', False to select sig='F'.
        None means select both.  See table below for explanation.
    calstate : bool
        Select the calibration state used to form the data.  True means select cal='T', False to select cal='F'.
        None means select both. See table below for explanation.
    scanrows : list-like
        the list of rows in `sdfits` corresponding to sigstate integrations
    calrows : dict
        dictionary containing with keys 'ON' and 'OFF' containing list of rows in `sdfits` corresponding to cal=T (ON) and cal=F (OFF) integrations for `scan`
    bintable : int
        the index for BINTABLE in `sdfits` containing the scans
    calibrate: bool
        whether or not to calibrate the data.  If `True`, the data will be (calon - caloff)*0.5, otherwise it will be SDFITS row data. Default:True
    smoothref: int
        the number of channels in the reference to boxcar smooth prior to calibration
    apply_flags : boolean, optional.  If True, apply flags before calibration.

    Notes
    -----
    How the total power and system temperature are calculated, depending on signal and reference state parameters:


     ======   =====   ===================================================================       ==================================
     CAL      SIG     RESULT                                                                    TSYS
     ======   =====   ===================================================================       ==================================
     None     None    data = 0.5* (REFCALON + REFCALOFF), regardless of sig state               use all CAL states, all SIG states
     None     True    data = 0.5* (REFCALON + REFCALOFF), where sig = 'T'                       use all CAL states, SIG='T'
     None     False   data = 0.5* (REFCALON + REFCALOFF), where sig = 'F'                       use all CAL states, SIG='F'
     True     None    data = REFCALON, regardless of sig state                                  use all CAL states, all SIG states
     False    None    data = REFCALOFF, regardless of sig state                                 use all CAL states, all SIG states
     True     True    data = REFCALON, where sig='T'                                            use all CAL states, SIG='T'
     True     False   data = REFCALON, where sig='F'                                            use all CAL states, SIG='F'
     False    True    data = REFCALOFF  where sig='T'                                           use all CAL states, SIG='T'
     False    False   data = REFCALOFF, where sig='F'                                           use all CAL states, SIG='F'
     ======   =====   ===================================================================       ==================================

    where `REFCALON` = integrations with `cal=T` and  `REFCALOFF` = integrations with `cal=F`.

    """

    def __init__(
        self,
        gbtfits,
        scan,
        sigstate,
        calstate,
        scanrows,
        calrows,
        bintable,
        calibrate=True,
        smoothref=1,
        apply_flags=False,
        observer_location=Observatory["GBT"],
    ):
        ScanBase.__init__(self, gbtfits)
        self._sdfits = gbtfits  # parent class
        self._scan = scan
        self._sigstate = sigstate
        self._calstate = calstate
        self._scanrows = scanrows
        self._smoothref = smoothref
        self._apply_flags = apply_flags
        if self._smoothref > 1:
            warnings.warn(f"TP smoothref={self._smoothref} not implemented yet")

        # @todo deal with data that crosses bintables
        if bintable is None:
            self._bintable_index = self._sdfits._find_bintable_and_row(self._scanrows[0])[0]
        else:
            self._bintable_index = bintable
        self._observer_location = observer_location
        df = self._sdfits._index
        df = df.iloc[scanrows]
        self._index = df
        self._set_if_fd(df)
        self._pols = uniq(df["PLNUM"])
        self._nint = 0
        self._npol = len(self._pols)
        self._timeaveraged = None
        self._polaveraged = None
        self._nrows = len(scanrows)
        self._tsys = None
        if False:
            self._npol = gbtfits.npol(self._bintable_index)  # @todo deal with bintable
            self._nint = gbtfits.nintegrations(self._bintable_index)
        self._calrows = calrows
        # all cal=T states where sig=sigstate
        self._refonrows = sorted(list(set(self._calrows["ON"]).intersection(set(self._scanrows))))
        # all cal=F states where sig=sigstate
        self._refoffrows = sorted(list(set(self._calrows["OFF"]).intersection(set(self._scanrows))))
        self._refcalon = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refonrows]
        self._refcaloff = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refoffrows]

        # Catch blank integrations.
        goodrows = find_nonblank_ints(self._refcalon, self._refcaloff)
        self._refcalon = self._refcalon[goodrows]
        self._refcaloff = self._refcaloff[goodrows]
        self._refonrows = [self._refonrows[i] for i in goodrows]
        self._refoffrows = [self._refoffrows[i] for i in goodrows]
        self._nrows = len(self._refonrows) + len(self._refoffrows)

        self._nchan = len(self._refcalon[0])
        self._calibrate = calibrate
        self._data = None
        if self._calibrate:
            self.calibrate()
        self.calc_tsys()
        self._validate_defaults()

    def calibrate(self):
        """Calibrate the data according to the CAL/SIG table above"""
        # the way the data are formed depend only on cal state
        # since we have downselected based on sig state in the constructor
        if self.calstate is None:
            self._data = (0.5 * (self._refcalon + self._refcaloff)).astype(float)
        elif self.calstate:
            self._data = self._refcalon.astype(float)
        elif self.calstate == False:
            self._data = self._refcaloff.astype(float)
        else:
            raise Exception(f"Unrecognized cal state {self.calstate}")  # should never happen

    @property
    def sigstate(self):
        """The requested signal state

        Returns
        -------
        bool
            True if signal state is on ('T' in the SDFITS header), False otherwise ('F')
        """
        return self._sigstate

    @property
    def calstate(self):
        """The requested calibration state

        Returns
        -------
        bool
            True if calibration state is on ('T' in the SDFITS header), False otherwise ('F')
        """
        return self._calstate

    def calc_tsys(self, **kwargs):
        """
        Calculate the system temperature array, according to table above.
        """
        kwargs_opts = {"verbose": False}
        kwargs_opts.update(kwargs)

        if False:
            if self.calstate is None:
                tcal = list(self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["TCAL"])
                nspect = len(tcal)
                # calon = self._refcalcon
                # caloff = self._refcaloff
            elif self.calstate:
                tcal = list(self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["TCAL"])
                nspect = len(tcal)
                # calon = self._refcalon
            elif self.calstate == False:
                pass
        self._tcal = list(self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["TCAL"])
        nspect = len(self._tcal)
        self._tsys = np.empty(nspect, dtype=float)  # should be same as len(calon)
        # allcal = self._refonrows.copy()
        # allcal.extend(self._refoffrows)
        # tcal = list(self._sdfits.index(self._bintable_index).iloc[sorted(allcal)]["TCAL"])
        # @todo this loop could be replaced with clever numpy
        if len(self._tcal) != nspect:
            raise Exception(f"TCAL length {len(tcal)} and number of spectra {nspect} don't match")
        for i in range(nspect):
            # tsys = mean_tsys(calon=calon[i], caloff=caloff[i], tcal=tcal[i])
            tsys = mean_tsys(calon=self._refcalon[i], caloff=self._refcaloff[i], tcal=self._tcal[i])
            self._tsys[i] = tsys

    @property
    def exposure(self):
        """The array of exposure (integration) times.  The value depends on the cal state:

            =====  ======================================
            CAL    EXPOSURE
            =====  ======================================
            None   :math:`t_{EXP,REFON} + t_{EXP,REFOFF}`
            True   :math:`t_{EXP,REFON}`
            False  :math:`t_{EXP,REFOFF}`
            =====  ======================================

        Returns
        -------
        exposure : `~numpy.ndarray`
            The exposure time in units of the EXPOSURE keyword in the SDFITS header
        """
        if self.calstate is None:
            exp_ref_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["EXPOSURE"].to_numpy()
            exp_ref_off = (
                self._sdfits.index(bintable=self._bintable_index).iloc[self._refoffrows]["EXPOSURE"].to_numpy()
            )
        elif self.calstate:
            exp_ref_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["EXPOSURE"].to_numpy()
            exp_ref_off = 0
        elif self.calstate == False:
            exp_ref_on = 0
            exp_ref_off = (
                self._sdfits.index(bintable=self._bintable_index).iloc[self._refoffrows]["EXPOSURE"].to_numpy()
            )

        self._exposure = exp_ref_on + exp_ref_off
        return self._exposure

    @property
    def delta_freq(self):
        r"""Get the array of channel frequency width. The value depends on the cal state:


        =====  ================================================================
        CAL     :math:`\Delta\nu`
        =====  ================================================================
        None    :math:`0.5 * ( \Delta\nu_{REFON}+ \Delta\nu_{REFOFF} )`
        True    :math:`\Delta\nu_{REFON}`
        False   :math:`\Delta\nu_{REFOFF}`
        =====  ================================================================


        Returns
        -------
        delta_freq: `~numpy.ndarray`
            The channel frequency width in units of the CDELT1 keyword in the SDFITS header
        """
        df_ref_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["CDELT1"].to_numpy()
        df_ref_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._refoffrows]["CDELT1"].to_numpy()
        if self.calstate is None:
            delta_freq = 0.5 * (df_ref_on + df_ref_off)
        elif self.calstate:
            delta_freq = df_ref_on
        elif self.calstate == False:
            delta_freq = df_ref_off
        self._delta_freq = delta_freq
        return self._delta_freq

    def tpmeta(self, i):
        ser = self._sdfits.index(bintable=self._bintable_index).iloc[self._scanrows[i]]
        meta = ser.dropna().to_dict()
        meta["TSYS"] = self._tsys[i]
        meta["EXPOSURE"] = self.exposure[i]
        meta["NAXIS1"] = len(self._data[i])
        if "CUNIT1" not in meta:
            meta["CUNIT1"] = "Hz"  # @todo this is in gbtfits.hdu[0].header['TUNIT11'] but is it always TUNIT11?
        meta["CUNIT2"] = "deg"  # is this always true?
        meta["CUNIT3"] = "deg"  # is this always true?
        restfrq = meta["RESTFREQ"]
        rfq = restfrq * u.Unit(meta["CUNIT1"])
        restfreq = rfq.to("Hz").value
        meta["RESTFRQ"] = restfreq  # WCS wants no E
        return meta

    def total_power(self, i):
        """Return the total power spectrum

        Parameters
        ----------
        i : int
            The index into the data array

        Returns
        -------
        spectrum : `~spectra.spectrum.Spectrum`
        """
        if not self._calibrate:
            raise Exception("You must calibrate first to get a total power spectrum")
        ser = self._sdfits.index(bintable=self._bintable_index).iloc[self._scanrows[i]]
        meta = ser.dropna().to_dict()
        meta["TSYS"] = self._tsys[i]
        meta["EXPOSURE"] = self.exposure[i]
        meta["NAXIS1"] = len(self._data[i])
        if "CUNIT1" not in meta:
            meta["CUNIT1"] = "Hz"  # @todo this is in gbtfits.hdu[0].header['TUNIT11'] but is it always TUNIT11?
        meta["CUNIT2"] = "deg"  # is this always true?
        meta["CUNIT3"] = "deg"  # is this always true?
        restfrq = meta["RESTFREQ"]
        rfq = restfrq * u.Unit(meta["CUNIT1"])
        restfreq = rfq.to("Hz").value
        meta["RESTFRQ"] = restfreq  # WCS wants no E

        s = Spectrum.make_spectrum(
            Masked(self._data[i] * u.ct, self._data[i].mask), meta=meta, observer_location=self._observer_location
        )
        s.merge_commentary(self)
        return s

    @log_call_to_history
    def timeaverage(self, weights="tsys"):
        r"""Compute the time-averaged spectrum for this set of scans.

        Parameters
        ----------
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'
        Returns
        -------
        spectrum : :class:`~spectra.spectrum.Spectrum`
            The time-averaged spectrum

        .. note::
           Data that are masked will have values set to zero.  This is a feature of `numpy.ma.average`. Data mask fill value is NaN (np.nan)
        """
        if self._npol > 1:
            raise Exception("Can't yet time average multiple polarizations")
        self._timeaveraged = deepcopy(self.total_power(0))
        if weights == "tsys":
            w = self.tsys_weight
        else:
            w = np.ones_like(self.tsys_weight)
        non_blanks = find_non_blanks(self._data)[0]
        self._timeaveraged._data = np.ma.average(self._data, axis=0, weights=w)
        self._timeaveraged._data.set_fill_value(np.nan)
        self._timeaveraged.meta["MEANTSYS"] = np.mean(self._tsys[non_blanks])
        self._timeaveraged.meta["WTTSYS"] = sq_weighted_avg(self._tsys[non_blanks], axis=0, weights=w[non_blanks])
        self._timeaveraged.meta["TSYS"] = self._timeaveraged.meta["WTTSYS"]
        self._timeaveraged.meta["EXPOSURE"] = self.exposure[non_blanks].sum()
        return self._timeaveraged


class PSScan(ScanBase):
    """GBT specific version of Position Switch Scan. A position switch scan object has
    one IF, one feed, and one or more polarizations.

    Parameters
    ----------
    gbtfits : `~fits.sdfitsload.SDFITSLoad`
        input SDFITSLoad object
    scans : dict
        dictionary with keys 'ON' and 'OFF' containing unique list of ON (signal) and OFF (reference) scan numbers NOTE: there should be one ON and one OFF, a pair
    scanrows : dict
        dictionary with keys 'ON' and 'OFF' containing the list of rows in `sdfits` corresponding to ON (signal) and OFF (reference) integrations
    calrows : dict
        dictionary containing with keys 'ON' and 'OFF' containing list of rows in `sdfits` corresponding to cal=T (ON) and cal=F (OFF) integrations.
    bintable : int
        the index for BINTABLE in `sdfits` containing the scans
    calibrate: bool
        whether or not to calibrate the data.  If true, data will be calibrated as TSYS*(ON-OFF)/OFF. Default: True
    smoothref: int
        the number of channels in the reference to boxcar smooth prior to calibration
    apply_flags : boolean, optional.  If True, apply flags before calibration.
    observer_location : `~astropy.coordinates.EarthLocation`
        Location of the observatory. See `~dysh.coordinates.Observatory`.
        This will be transformed to `~astropy.coordinates.ITRS` using the time of
        observation DATE-OBS or MJD-OBS in
        the SDFITS header.  The default is the location of the GBT.
    """

    def __init__(
        self,
        gbtfits,
        scans,
        scanrows,
        calrows,
        bintable,
        calibrate=True,
        smoothref=1,
        apply_flags=False,
        observer_location=Observatory["GBT"],
    ):
        ScanBase.__init__(self, gbtfits)
        # The rows of the original bintable corresponding to ON (sig) and OFF (reg)
        # self._scans = scans
        # self._history = deepcopy(gbtfits._history)
        self._scan = scans["ON"]
        self._scanrows = scanrows
        self._nrows = len(self._scanrows["ON"])
        self._smoothref = smoothref
        self._apply_flags = apply_flags

        # calrows perhaps not needed as input since we can get it from gbtfits object?
        # calrows['ON'] are rows with noise diode was on, regardless of sig or ref
        # calrows['OFF'] are rows with noise diode was off, regardless of sig or ref
        self._calrows = calrows
        # @todo deal with data that crosses bintables
        if bintable is None:
            self._bintable_index = gbtfits._find_bintable_and_row(self._scanrows["ON"][0])[0]
        else:
            self._bintable_index = bintable
        self._observer_location = observer_location
        # df = selection.iloc[scanrows["ON"]]
        df = self._sdfits._index.iloc[scanrows["ON"]]
        self._set_if_fd(df)
        self._pols = uniq(df["PLNUM"])
        self._npol = len(self._pols)
        if False:
            self._nint = gbtfits.nintegrations(self._bintable_index)
        # so quick with slicing!
        self._sigonrows = sorted(list(set(self._calrows["ON"]).intersection(set(self._scanrows["ON"]))))
        self._sigoffrows = sorted(list(set(self._calrows["OFF"]).intersection(set(self._scanrows["ON"]))))
        self._refonrows = sorted(list(set(self._calrows["ON"]).intersection(set(self._scanrows["OFF"]))))
        self._refoffrows = sorted(list(set(self._calrows["OFF"]).intersection(set(self._scanrows["OFF"]))))
        self._sigcalon = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._sigonrows]
        self._sigcaloff = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._sigoffrows]
        self._refcalon = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refonrows]
        self._refcaloff = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refoffrows]

        # Catch blank integrations.
        goodrows = find_nonblank_ints(self._sigcaloff, self._refcaloff, self._sigcalon, self._refcalon)
        self._refcalon = self._refcalon[goodrows]
        self._refcaloff = self._refcaloff[goodrows]
        self._refonrows = [self._refonrows[i] for i in goodrows]
        self._refoffrows = [self._refoffrows[i] for i in goodrows]
        self._sigcalon = self._sigcalon[goodrows]
        self._sigcaloff = self._sigcaloff[goodrows]
        self._sigonrows = [self._sigonrows[i] for i in goodrows]
        self._sigoffrows = [self._sigoffrows[i] for i in goodrows]
        # Update number of rows after removing blanks.
        nsigrows = len(self._sigonrows) + len(self._sigoffrows)
        self._nrows = nsigrows

        self._nchan = gbtfits.nchan(self._bintable_index)
        self._tsys = None
        self._exposure = None
        self._calibrated = None
        self._calibrate = calibrate
        self._make_meta(self._sigonrows)
        if self._calibrate:
            self.calibrate()
        self._validate_defaults()

    # @property
    # def scans(self):
    #     """The dictionary of the ON and OFF scan numbers in the PSScan.
    #
    #     Returns
    #     -------
    #     scans : dict
    #         The scan number dictionary
    #
    #     """
    #      return self._scans

    # @todo something clever
    # self._calibrated_spectrum = Spectrum(self._calibrated,...) [assuming same spectral axis]
    def calibrated(self, i):
        """Return the calibrated Spectrum.

        Parameters
        ----------
        i : int
            The index into the calibrated array

        Returns
        -------
        spectrum : `~spectra.spectrum.Spectrum`
        """
        # @todo suppress astropy INFO message "overwriting Masked Quantity's current mask with specified mask."
        s = Spectrum.make_spectrum(
            Masked(self._calibrated[i] * u.K, self._calibrated[i].mask),
            meta=self.meta[i],
            observer_location=self._observer_location,
        )
        s.merge_commentary(self)
        return s

    def calibrate(self, **kwargs):
        """
        Position switch calibration, following equations 1 and 2 in the GBTIDL calibration manual
        """
        kwargs_opts = {"verbose": False}
        kwargs_opts.update(kwargs)
        if self._smoothref > 1 and kwargs_opts["verbose"]:
            print(f"PS smoothref={self._smoothref}")

        self._status = 1
        nspect = self.nrows // 2
        self._calibrated = np.ma.empty((nspect, self._nchan), dtype="d")
        self._tsys = np.empty(nspect, dtype="d")
        self._exposure = np.empty(nspect, dtype="d")
        tcal = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["TCAL"].to_numpy()
        # @todo  this loop could be replaced with clever numpy
        if len(tcal) != nspect:
            raise Exception(f"TCAL length {len(tcal)} and number of spectra {nspect} don't match")
        for i in range(nspect):
            tsys = mean_tsys(calon=self._refcalon[i], caloff=self._refcaloff[i], tcal=tcal[i])
            sig = 0.5 * (self._sigcalon[i] + self._sigcaloff[i])
            ref = 0.5 * (self._refcalon[i] + self._refcaloff[i])
            if self._smoothref > 1:
                ref = core.smooth(ref, "boxcar", self._smoothref)
            self._calibrated[i] = tsys * (sig - ref) / ref
            self._tsys[i] = tsys
            self._exposure[i] = self.exposure[i]
        logger.debug(f"Calibrated {nspect} spectra")
        self._add_calibration_meta()

    # tip o' the hat to Pedro S. for exposure and delta_freq
    @property
    def exposure(self):
        """The array of exposure (integration) times

        exposure = [ 0.5*(exp_ref_on + exp_ref_off) + 0.5*(exp_sig_on + exp_sig_off) ] / 2

        Returns
        -------
        exposure : ~numpy.ndarray
            The exposure time in units of the EXPOSURE keyword in the SDFITS header
        """
        exp_ref_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["EXPOSURE"].to_numpy()
        exp_ref_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._refoffrows]["EXPOSURE"].to_numpy()
        exp_sig_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigonrows]["EXPOSURE"].to_numpy()
        exp_sig_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigoffrows]["EXPOSURE"].to_numpy()
        exp_ref = exp_ref_on + exp_ref_off
        exp_sig = exp_sig_on + exp_sig_off
        if self._smoothref > 1:
            nsmooth = self._smoothref
        else:
            nsmooth = 1.0
        exposure = exp_sig * exp_ref * nsmooth / (exp_sig + exp_ref * nsmooth)
        return exposure

    @property
    def delta_freq(self):
        """Get the array of channel frequency width

        df = [ 0.5*(df_ref_on + df_ref_off) + 0.5*(df_sig_on + df_sig_off) ] / 2

        Returns
        -------
             delta_freq: ~numpy.ndarray
                 The channel frequency width in units of the CDELT1 keyword in the SDFITS header
        """
        df_ref_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["CDELT1"].to_numpy()
        df_ref_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._refoffrows]["CDELT1"].to_numpy()
        df_sig_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigonrows]["CDELT1"].to_numpy()
        df_sig_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigoffrows]["CDELT1"].to_numpy()
        df_ref = 0.5 * (df_ref_on + df_ref_off)
        df_sig = 0.5 * (df_sig_on + df_sig_off)
        delta_freq = 0.5 * (df_ref + df_sig)
        return delta_freq

    @log_call_to_history
    def timeaverage(self, weights="tsys"):
        r"""Compute the time-averaged spectrum for this set of scans.

        Parameters
        ----------
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'

        Returns
        -------
        spectrum : :class:`~spectra.spectrum.Spectrum`
            The time-averaged spectrum

        .. note::
           Data that are masked will have values set to zero.  This is a feature of `numpy.ma.average`. Data mask fill value is NaN (np.nan)
        """
        if self._calibrated is None or len(self._calibrated) == 0:
            raise Exception("You can't time average before calibration.")
        if self._npol > 1:
            raise Exception("Can't yet time average multiple polarizations")
        self._timeaveraged = deepcopy(self.calibrated(0))  # ._copy()
        data = self._calibrated
        if weights == "tsys":
            w = self.tsys_weight
        else:
            w = np.ones_like(self.tsys_weight)
        self._timeaveraged._data = np.ma.average(data, axis=0, weights=w)
        self._timeaveraged._data.set_fill_value(np.nan)
        non_blanks = find_non_blanks(data)
        self._timeaveraged.meta["MEANTSYS"] = np.mean(self._tsys[non_blanks])
        self._timeaveraged.meta["WTTSYS"] = sq_weighted_avg(self._tsys[non_blanks], axis=0, weights=w[non_blanks])
        self._timeaveraged.meta["EXPOSURE"] = np.sum(self._exposure[non_blanks])
        self._timeaveraged.meta["TSYS"] = self._timeaveraged.meta["WTTSYS"]
        self._timeaveraged._history = self._history
        self._timeaveraged._observer_location = self._observer_location
        return self._timeaveraged


class NodScan(ScanBase):
    """GBT specific version of Nodding Scan. A nod scan object has
    one IF, two feeds, and one or more polarizations.

    Parameters
    ----------
    gbtfits : `~fits.sdfitsload.SDFITSLoad`
        input SDFITSLoad object
    scan : dict
        dictionary with keys 'ON' and 'OFF' containing unique list of ON (signal) and OFF (reference) scan numbers
        NOTE: there should be one ON and one OFF, a pair. There should be at least two beams (the nodding beams)
        which will be resp. on source in each scan.
    beam1: bool
        Is this scan BEAM1 or BEAM2?  BEAM1 is defined as being on source in the first scan of the pair, BEAM2 in the second of the pair
    scanrows : dict
        dictionary with keys 'ON' and 'OFF' containing the list of rows in `sdfits` corresponding to ON (signal) and OFF (reference) integrations
    calrows : dict
        dictionary containing with keys 'ON' and 'OFF' containing list of rows in `sdfits` corresponding to cal=T (ON) and cal=F (OFF) integrations.
    bintable : int
        the index for BINTABLE in `sdfits` containing the scans
    calibrate: bool
        whether or not to calibrate the data.  If true, data will be calibrated as TSYS*(ON-OFF)/OFF.
        Default: True
    smoothref: int
        the number of channels in the reference to boxcar smooth prior to calibration (if applicable)
    apply_flags : boolean, optional.  If True, apply flags before calibration.
    observer_location : `~astropy.coordinates.EarthLocation`
        Location of the observatory. See `~dysh.coordinates.Observatory`.
        This will be transformed to `~astropy.coordinates.ITRS` using the time of
        observation DATE-OBS or MJD-OBS in
        the SDFITS header.  The default is the location of the GBT.
    """

    def __init__(
        self,
        gbtfits,
        scan,
        beam1,
        scanrows,
        calrows,
        bintable,
        calibrate=True,
        smoothref=1,
        apply_flags=False,
        observer_location=Observatory["GBT"],
    ):
        ScanBase.__init__(self, gbtfits)
        self._scan = scan["ON"]
        self._scanrows = scanrows
        self._nrows = len(self._scanrows["ON"])
        self._smoothref = smoothref
        self._apply_flags = apply_flags
        self._beam1 = beam1

        # @todo   allow having no calrow where noise diode was not fired

        # calrows perhaps not needed as input since we can get it from gbtfits object?
        # calrows['ON'] are rows with noise diode was on, regardless of sig or ref
        # calrows['OFF'] are rows with noise diode was off, regardless of sig or ref
        self._calrows = calrows
        # @todo deal with data that crosses bintables
        if bintable is None:
            self._bintable_index = gbtfits._find_bintable_and_row(self._scanrows["ON"][0])[0]
        else:
            self._bintable_index = bintable
        self._observer_location = observer_location
        # df = selection.iloc[scanrows["ON"]]
        df = self._sdfits._index.iloc[scanrows["ON"]]
        self._set_if_fd(df)
        self._pols = uniq(df["PLNUM"])
        self._npol = len(self._pols)
        if False:
            self._nint = gbtfits.nintegrations(self._bintable_index)
        # so quick with slicing!
        self._sigonrows = sorted(list(set(self._calrows["ON"]).intersection(set(self._scanrows["ON"]))))
        self._sigoffrows = sorted(list(set(self._calrows["OFF"]).intersection(set(self._scanrows["ON"]))))
        self._refonrows = sorted(list(set(self._calrows["ON"]).intersection(set(self._scanrows["OFF"]))))
        self._refoffrows = sorted(list(set(self._calrows["OFF"]).intersection(set(self._scanrows["OFF"]))))
        if beam1:
            self._sigcalon = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._sigonrows]
            self._sigcaloff = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._sigoffrows]
            self._refcalon = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refonrows]
            self._refcaloff = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refoffrows]
        else:
            self._sigcalon = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refonrows]
            self._sigcaloff = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refoffrows]
            self._refcalon = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._sigonrows]
            self._refcaloff = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._sigoffrows]

        # Catch blank integrations.
        goodrows = find_nonblank_ints(self._sigcaloff, self._refcaloff, self._sigcalon, self._refcalon)
        self._refcalon = self._refcalon[goodrows]
        self._refcaloff = self._refcaloff[goodrows]
        self._refonrows = [self._refonrows[i] for i in goodrows]
        self._refoffrows = [self._refoffrows[i] for i in goodrows]
        self._sigcalon = self._sigcalon[goodrows]
        self._sigcaloff = self._sigcaloff[goodrows]
        self._sigonrows = [self._sigonrows[i] for i in goodrows]
        self._sigoffrows = [self._sigoffrows[i] for i in goodrows]
        # Update number of rows after removing blanks.
        nsigrows = len(self._sigonrows) + len(self._sigoffrows)
        self._nrows = nsigrows

        self._nchan = len(self._sigcalon[0])
        self._tsys = None
        self._exposure = None
        self._calibrated = None
        self._calibrate = calibrate
        self._make_meta(self._sigonrows)
        if self._calibrate:
            self.calibrate()
        self._validate_defaults()

    # @property
    # def scans(self):
    #     """The dictionary of the ON and OFF scan numbers in the PSScan.
    #
    #     Returns
    #     -------
    #     scans : dict
    #         The scan number dictionary
    #
    #     """
    #      return self._scans

    # @todo something clever
    # self._calibrated_spectrum = Spectrum(self._calibrated,...) [assuming same spectral axis]
    def calibrated(self, i):
        """Return the calibrated Spectrum.

        Parameters
        ----------
        i : int
            The index into the calibrated array

        Returns
        -------
        spectrum : `~spectra.spectrum.Spectrum`
        """
        s = Spectrum.make_spectrum(
            Masked(self._calibrated[i] * u.K, self._calibrated[i].mask),
            meta=self.meta[i],
            observer_location=self._observer_location,
        )
        s.merge_commentary(self)
        return s

    def calibrate(self, **kwargs):
        """
        Position switch calibration, following equations 1 and 2 in the GBTIDL calibration manual
        """
        kwargs_opts = {"verbose": False}
        kwargs_opts.update(kwargs)
        if self._smoothref > 1 and kwargs_opts["verbose"]:
            print(f"PS smoothref={self._smoothref}")

        self._status = 1
        nspect = self.nrows // 2
        self._calibrated = np.ma.empty((nspect, self._nchan), dtype="d")
        self._tsys = np.empty(nspect, dtype="d")
        self._exposure = np.empty(nspect, dtype="d")
        tcal = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["TCAL"].to_numpy()
        # @todo  this loop could be replaced with clever numpy
        if len(tcal) != nspect:
            raise Exception(f"TCAL length {len(tcal)} and number of spectra {nspect} don't match")
        for i in range(nspect):
            tsys = mean_tsys(calon=self._refcalon[i], caloff=self._refcaloff[i], tcal=tcal[i])
            sig = 0.5 * (self._sigcalon[i] + self._sigcaloff[i])
            ref = 0.5 * (self._refcalon[i] + self._refcaloff[i])
            if self._smoothref > 1:
                ref = core.smooth(ref, "boxcar", self._smoothref)
            self._calibrated[i] = tsys * (sig - ref) / ref
            self._tsys[i] = tsys
            self._exposure[i] = self.exposure[i]
        logger.debug(f"Calibrated {nspect} spectra")
        self._add_calibration_meta()

    # tip o' the hat to Pedro S. for exposure and delta_freq
    @property
    def exposure(self):
        """The array of exposure (integration) times

        exposure = [ 0.5*(exp_ref_on + exp_ref_off) + 0.5*(exp_sig_on + exp_sig_off) ] / 2

        Returns
        -------
        exposure : ~numpy.ndarray
            The exposure time in units of the EXPOSURE keyword in the SDFITS header
        """
        exp_ref_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["EXPOSURE"].to_numpy()
        exp_ref_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._refoffrows]["EXPOSURE"].to_numpy()
        exp_sig_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigonrows]["EXPOSURE"].to_numpy()
        exp_sig_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigoffrows]["EXPOSURE"].to_numpy()
        exp_ref = exp_ref_on + exp_ref_off
        exp_sig = exp_sig_on + exp_sig_off
        if self._smoothref > 1:
            nsmooth = self._smoothref
        else:
            nsmooth = 1.0
        exposure = exp_sig * exp_ref * nsmooth / (exp_sig + exp_ref * nsmooth)
        return exposure

    @property
    def delta_freq(self):
        """Get the array of channel frequency width

        df = [ 0.5*(df_ref_on + df_ref_off) + 0.5*(df_sig_on + df_sig_off) ] / 2

        Returns
        -------
             delta_freq: ~numpy.ndarray
                 The channel frequency width in units of the CDELT1 keyword in the SDFITS header
        """
        df_ref_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["CDELT1"].to_numpy()
        df_ref_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._refoffrows]["CDELT1"].to_numpy()
        df_sig_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigonrows]["CDELT1"].to_numpy()
        df_sig_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigoffrows]["CDELT1"].to_numpy()
        df_ref = 0.5 * (df_ref_on + df_ref_off)
        df_sig = 0.5 * (df_sig_on + df_sig_off)
        delta_freq = 0.5 * (df_ref + df_sig)
        return delta_freq

    @log_call_to_history
    def timeaverage(self, weights="tsys"):
        r"""Compute the time-averaged spectrum for this set of scans.

        Parameters
        ----------
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'
        Returns
        -------
        spectrum : :class:`~spectra.spectrum.Spectrum`
            The time-averaged spectrum

        .. note::
           Data that are masked will have values set to zero.  This is a feature of `numpy.ma.average`. Data mask fill value is NaN (np.nan)
        """
        if self._calibrated is None or len(self._calibrated) == 0:
            raise Exception("You can't time average before calibration.")
        if self._npol > 1:
            raise Exception("Can't yet time average multiple polarizations")
        self._timeaveraged = deepcopy(self.calibrated(0))
        data = self._calibrated
        if weights == "tsys":
            w = self.tsys_weight
        else:
            w = np.ones_like(self.tsys_weight)
        self._timeaveraged._data = np.ma.average(data, axis=0, weights=w)
        self._timeaveraged._data.set_fill_value(np.nan)
        non_blanks = find_non_blanks(data)
        self._timeaveraged.meta["MEANTSYS"] = np.mean(self._tsys[non_blanks])
        self._timeaveraged.meta["WTTSYS"] = sq_weighted_avg(self._tsys[non_blanks], axis=0, weights=w[non_blanks])
        self._timeaveraged.meta["EXPOSURE"] = np.sum(self._exposure[non_blanks])
        self._timeaveraged.meta["TSYS"] = self._timeaveraged.meta["WTTSYS"]
        self._timeaveraged._history = self._history
        return self._timeaveraged


class FSScan(ScanBase):
    """GBT specific version of Frequency Switch Scan

    Parameters
    ----------

    gbtfits : `~fits.sdfitsload.SDFITSLoad`
        input SDFITSLoad object
    scan : int
        Scan number that contains integrations with a series of sig/ref and calon/caloff states.
    sigrows :dict
        Dictionary containing with keys 'ON' and 'OFF' containing list of rows in `sdfits`
        corresponding to sig=T (ON) and sig=F (OFF) integrations.
    calrows : dict
        Dictionary containing with keys 'ON' and 'OFF' containing list of rows in `sdfits`
        corresponding to cal=T (ON) and cal=F (OFF) integrations.
    bintable : int
        The index for BINTABLE in `sdfits` containing the scans.
    calibrate : bool
        Whether or not to calibrate the data.  If true, data will be calibrated as TSYS*(ON-OFF)/OFF.
        Default: True
    fold : bool
        Whether or not to fold the spectrum. Default: True
    shift_method : str
        Method to use when shifting the spectra for folding. One of 'fft' or 'interpolate'.
        'fft' uses a phase shift in the time domain. 'interpolate' interpolates the signal. Default: 'fft'
    use_sig : bool
        Whether to use the sig as the sig, or the ref as the sig. Default: True
    smoothref: int
        The number of channels in the reference to boxcar smooth prior to calibration.
    apply_flags : boolean, optional.  If True, apply flags before calibration.
    observer_location : `~astropy.coordinates.EarthLocation`
        Location of the observatory. See `~dysh.coordinates.Observatory`.
        This will be transformed to `~astropy.coordinates.ITRS` using the time of
        observation DATE-OBS or MJD-OBS in the SDFITS header.  The default is the location of the GBT.
    """

    def __init__(
        self,
        gbtfits,
        scan,
        sigrows,
        calrows,
        bintable,
        calibrate=True,
        fold=True,
        shift_method="fft",
        use_sig=True,
        smoothref=1,
        apply_flags=False,
        observer_location=Observatory["GBT"],
        debug=False,
    ):
        ScanBase.__init__(self, gbtfits)
        # The rows of the original bintable corresponding to ON (sig) and OFF (reg)
        self._scan = scan  # for FS everything is an "ON"
        self._sigrows = sigrows  # dict with "ON" and "OFF"
        self._calrows = calrows  # dict with "ON" and "OFF"
        self._folded = False
        self._use_sig = use_sig
        self._smoothref = smoothref
        if self._smoothref > 1:
            print(f"FS smoothref={self._smoothref} not implemented yet")
        self._apply_flags = apply_flags
        self._sigonrows = sorted(list(set(self._calrows["ON"]).intersection(set(self._sigrows["ON"]))))
        self._sigoffrows = sorted(list(set(self._calrows["OFF"]).intersection(set(self._sigrows["ON"]))))
        self._refonrows = sorted(list(set(self._calrows["ON"]).intersection(set(self._sigrows["OFF"]))))
        self._refoffrows = sorted(list(set(self._calrows["OFF"]).intersection(set(self._sigrows["OFF"]))))

        self._debug = debug

        if self._debug:
            logger.debug("---------------------------------------------------")
            logger.debug("FSSCAN: ")
            logger.debug(f"SigOff {self._sigoffrows}")
            logger.debug(f"SigOn {self._sigonrows}")
            logger.debug(f"RefOff {self._refoffrows}")
            logger.debug(f"RefOn {self._refonrows}")

        nsigrows = len(self._sigonrows) + len(self._sigoffrows)
        nrefrows = len(self._refonrows) + len(self._refoffrows)
        if nsigrows != nrefrows:
            raise Exception("Number of sig rows does not match ref rows. Dangerous to proceed")
        logger.debug(f"sigonrows {nsigrows}, {self._sigonrows}")
        self._nrows = nsigrows

        a_scanrow = self._sigonrows[0]

        # @todo deal with data that crosses bintables
        if bintable is None:
            self._bintable_index = gbtfits._find_bintable_and_row(a_scanrow)[0]
        else:
            self._bintable_index = bintable
        if self._debug:
            logger.debug(f"bintable index is {self._bintable_index}")
        self._observer_location = observer_location
        self._scanrows = list(set(self._calrows["ON"])) + list(set(self._calrows["OFF"]))

        df = self._sdfits._index.iloc[self._scanrows]
        if self._debug:
            logger.debug(f"{len(df) = }")
        self._set_if_fd(df)
        self._pols = uniq(df["PLNUM"])
        if self._debug:
            logger.debug(f"FSSCAN #pol = {self._pols}")
        self._npol = len(self._pols)
        if False:
            self._nint = gbtfits.nintegrations(self._bintable_index)
        # @todo use gbtfits.velocity_convention(veldef,velframe)
        # so quick with slicing!

        self._sigcalon = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._sigonrows]
        self._sigcaloff = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._sigoffrows]
        self._refcalon = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refonrows]
        self._refcaloff = gbtfits.rawspectra(self._bintable_index, setmask=apply_flags)[self._refoffrows]

        # Catch blank integrations.
        goodrows = find_nonblank_ints(self._sigcaloff, self._refcaloff, self._sigcalon, self._refcalon)
        self._refcalon = self._refcalon[goodrows]
        self._refcaloff = self._refcaloff[goodrows]
        self._refonrows = [self._refonrows[i] for i in goodrows]
        self._refoffrows = [self._refoffrows[i] for i in goodrows]
        self._sigcalon = self._sigcalon[goodrows]
        self._sigcaloff = self._sigcaloff[goodrows]
        self._sigonrows = [self._sigonrows[i] for i in goodrows]
        self._sigoffrows = [self._sigoffrows[i] for i in goodrows]
        # Update number of rows after removing blanks.
        nsigrows = len(self._sigonrows) + len(self._sigoffrows)
        self._nrows = nsigrows

        self._nchan = len(self._sigcalon[0])
        self._tsys = None
        self._exposure = None
        self._calibrated = None
        self._calibrate = calibrate
        self._make_meta(self._sigonrows)
        if self._calibrate:
            self.calibrate(fold=fold, shift_method=shift_method)
        if self._debug:
            logger.debug("---------------------------------------------------")
        self._validate_defaults()

    @property
    def folded(self):
        """
        Has the FSscan been folded?

        Returns
        -------
        boolean
            True if the signal and reference integrations have been folded. False if not.
        """
        return self._folded

    def calibrated(self, i):
        """Return the calibrated Spectrum of this FSscan

        Parameters
        ----------
        i : int
            The index into the calibrated array

        Returns
        -------
        spectrum : `~spectra.spectrum.Spectrum`
        """
        s = Spectrum.make_spectrum(
            Masked(self._calibrated[i] * u.K, self._calibrated[i].mask),
            meta=self.meta[i],
            observer_location=self._observer_location,
        )
        s.merge_commentary(self)
        return s

    def calibrate(self, **kwargs):
        """
        Frequency switch calibration, following equations ...
        fold=True or fold=False is required
        """
        if self._debug:
            logger.debug(f'FOLD={kwargs["fold"]}')
            logger.debug(f'METHOD={kwargs["shift_method"]}')

        # some helper functions, courtesy proto_getfs.py
        def channel_to_frequency(crval1, crpix1, cdelt1, vframe, nchan, nint, ndim=1):
            """ """

            # Compute the correction factor.
            beta = (vframe * u.m / u.s) / ac.c
            vcorr = np.sqrt((1.0 + beta) / (1.0 - beta))

            # The +1 is to start counting from 1.
            indx = np.arange(nchan) + 1
            if ndim == 1:
                freq = crval1 + cdelt1 * (indx - crpix1)
                freq *= vcorr
            elif ndim == 2:
                indx = np.tile(indx, (nint, 1))
                freq = crval1[:, np.newaxis] + cdelt1[:, np.newaxis] * (indx - crpix1[:, np.newaxis])
                freq *= vcorr[:, np.newaxis]

            return freq

        def index_frequency(df):
            """
            Create a frequency axis from an index.
            This assumes all entries in the index have the same number of channels.
            """
            # Could you do this with gbtfits.getspec(row).spectral_axis?

            ndim = len(df.shape)
            nint = df.shape[0]

            if ndim == 1:
                nchan = np.array([int(df["TDIM7"][1:-1].split(",")[0])])
            else:
                nchan = np.array([int(df["TDIM7"].iloc[i][1:-1].split(",")[0]) for i in range(len(df))])

            crval1 = df["CRVAL1"]
            crpix1 = df["CRPIX1"]
            cdelt1 = df["CDELT1"]
            vframe = df["VFRAME"]  # Use the velocity frame requested by the user.

            if ndim == 2:
                crval1 = crval1.to_numpy()
                crpix1 = crpix1.to_numpy()
                cdelt1 = cdelt1.to_numpy()
                vframe = vframe.to_numpy()

            freq = channel_to_frequency(crval1, crpix1, cdelt1, vframe, nchan[0], nint, ndim=ndim)

            # Apply units.
            try:
                cunit1 = u.Unit(df["CUNIT1"])
                if ndim == 2:
                    cunit1 = cunit[0]  #  @todo undefined cunit[]
            except KeyError:
                cunit1 = u.Hz

            return freq * cunit1

        def do_total_power(no_cal, cal, tcal):
            """ """

        def vec_mean_tsys(on, off, tcal):
            """
            mean_tsys implements this, albeit only in 1D
            """
            pass

        def do_sig_ref(sig, ref, tsys, smooth=False):
            """
            smooth=True would implement smoothing the reference (or something)
            """
            return (sig - ref) / ref * tsys

        def do_fold(sig, ref, sig_freq, ref_freq, remove_wrap=False, shift_method="fft"):
            """ """
            chan_shift = (sig_freq[0] - ref_freq[0]) / np.abs(np.diff(sig_freq)).mean()
            logger.debug(f"do_fold: {sig_freq[0]}, {ref_freq[0]},{chan_shift}")
            ref_shift = core.data_shift(ref, chan_shift, remove_wrap=remove_wrap, method=shift_method)
            # @todo weights
            avg = (sig + ref_shift) / 2
            return avg

        kwargs_opts = {"verbose": False}
        kwargs_opts.update(kwargs)
        _fold = kwargs.get("fold", False)
        _mode = 1  # 1: keep the sig    else: keep the ref     (not externally supported)
        nspect = self.nrows // 2
        self._calibrated = np.ma.empty((nspect, self._nchan), dtype="d")
        self._tsys = np.empty(nspect, dtype="d")
        self._exposure = np.empty(nspect, dtype="d")
        #
        sig_freq = self._sigcalon[0]
        df_sig = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigonrows]
        df_ref = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]
        logger.debug(f"df_sig {type(df_sig)} len(df_sig)")
        sig_freq = index_frequency(df_sig)
        ref_freq = index_frequency(df_ref)
        chan_shift = abs(sig_freq[0, 0] - ref_freq[0, 0]) / np.abs(np.diff(sig_freq)).mean()
        logger.debug(f"FS: shift={chan_shift:g}  nchan={self._nchan:g}")

        #  tcal is the same for REF and SIG, and the same for all integrations actually.
        tcal = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigonrows]["TCAL"].to_numpy()
        logger.debug(f"TCAL: {len(tcal)} {tcal[0]}")
        if len(tcal) != nspect:
            raise Exception(f"TCAL length {len(tcal)} and number of spectra {nspect} don't match")
        # @todo   the nspect loop could be replaced with clever numpy?
        for i in range(nspect):
            tsys_sig = mean_tsys(calon=self._sigcalon[i], caloff=self._sigcaloff[i], tcal=tcal[i])
            tsys_ref = mean_tsys(calon=self._refcalon[i], caloff=self._refcaloff[i], tcal=tcal[i])
            if i == 0:
                logger.debug(f"Tsys(sig/ref)[0]={tsys_sig} / {tsys_ref}")
            tp_sig = 0.5 * (self._sigcalon[i] + self._sigcaloff[i])
            tp_ref = 0.5 * (self._refcalon[i] + self._refcaloff[i])
            #
            cal_sig = do_sig_ref(tp_sig, tp_ref, tsys_ref)
            cal_ref = do_sig_ref(tp_ref, tp_sig, tsys_sig)
            #
            if _fold:
                cal_sig_fold = do_fold(cal_sig, cal_ref, sig_freq[i], ref_freq[i], shift_method=kwargs["shift_method"])
                cal_ref_fold = do_fold(cal_ref, cal_sig, ref_freq[i], sig_freq[i], shift_method=kwargs["shift_method"])
                self._folded = True
                if self._use_sig:
                    self._calibrated[i] = cal_sig_fold
                    self._tsys[i] = tsys_ref
                else:
                    self._calibrated[i] = cal_ref_fold
                    self._tsys[i] = tsys_sig
                self._exposure[i] = 2 * self.exposure[i]  # @todo
            else:
                if self._use_sig:
                    self._calibrated[i] = cal_sig
                    self._tsys[i] = tsys_ref
                else:
                    self._calibrated[i] = cal_ref
                    self._tsys[i] = tsys_sig
                self._exposure[i] = self.exposure[i]

        self._add_calibration_meta()
        logger.debug(f"Calibrated {nspect} spectra with fold={_fold} and use_sig={self._use_sig}")

    # tip o' the hat to Pedro S. for exposure and delta_freq
    @property
    def exposure(self):
        """The array of exposure (integration) times for FSscan

        exposure = [ 0.5*(exp_ref_on + exp_ref_off) + 0.5*(exp_sig_on + exp_sig_off) ] / 2

        Returns
        -------
        exposure : ~numpy.ndarray
            The exposure time in units of the EXPOSURE keyword in the SDFITS header
        """
        exp_ref_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["EXPOSURE"].to_numpy()
        exp_ref_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._refoffrows]["EXPOSURE"].to_numpy()
        exp_sig_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigonrows]["EXPOSURE"].to_numpy()
        exp_sig_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigoffrows]["EXPOSURE"].to_numpy()
        exp_ref = exp_ref_on + exp_ref_off
        exp_sig = exp_sig_on + exp_sig_off
        if self._smoothref > 1:
            nsmooth = self._smoothref
        else:
            nsmooth = 1.0
        self._exposure = exp_sig * exp_ref * nsmooth / (exp_sig + exp_ref * nsmooth)
        return self._exposure

    @property
    def delta_freq(self):
        """Get the array of channel frequency width

        df = [ 0.5*(df_ref_on + df_ref_off) + 0.5*(df_sig_on + df_sig_off) ] / 2

        Returns
        -------
             delta_freq: ~numpy.ndarray
                 The channel frequency width in units of the CDELT1 keyword in the SDFITS header
        """
        df_ref_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._refonrows]["CDELT1"].to_numpy()
        df_ref_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._refoffrows]["CDELT1"].to_numpy()
        df_sig_on = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigonrows]["CDELT1"].to_numpy()
        df_sig_off = self._sdfits.index(bintable=self._bintable_index).iloc[self._sigoffrows]["CDELT1"].to_numpy()
        df_ref = 0.5 * (df_ref_on + df_ref_off)
        df_sig = 0.5 * (df_sig_on + df_sig_off)
        self._delta_freq = 0.5 * (df_ref + df_sig)
        return self._delta_freq

    def timeaverage(self, weights="tsys"):
        r"""Compute the time-averaged spectrum for this set of FSscans.

        Parameters
        ----------
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'
        Returns
        -------
        spectrum : :class:`~spectra.spectrum.Spectrum`
            The time-averaged spectrum

        .. note::
           Data that are masked will have values set to zero.  This is a feature of `numpy.ma.average`. Data mask fill value is NaN (np.nan)
        """
        if self._calibrated is None or len(self._calibrated) == 0:
            raise Exception("You can't time average before calibration.")
        if self._npol > 1:
            raise Exception("Can't yet time average multiple polarizations %d" % self._npol)
        self._timeaveraged = deepcopy(self.calibrated(0))
        data = self._calibrated
        if weights == "tsys":
            w = self.tsys_weight
        else:
            w = np.ones_like(self.tsys_weight)
        self._timeaveraged._data = np.ma.average(data, axis=0, weights=w)
        self._timeaveraged._data.set_fill_value(np.nan)
        non_blanks = find_non_blanks(data)
        self._timeaveraged.meta["MEANTSYS"] = np.mean(self._tsys[non_blanks])
        self._timeaveraged.meta["WTTSYS"] = sq_weighted_avg(self._tsys[non_blanks], axis=0, weights=w[non_blanks])
        self._timeaveraged.meta["EXPOSURE"] = np.sum(self._exposure[non_blanks])
        self._timeaveraged.meta["TSYS"] = self._timeaveraged.meta["WTTSYS"]
        return self._timeaveraged


class SubBeamNodScan(ScanBase):
    r"""
    Parameters
    ----------
    sigtp:  list of ~spectra.scan.TPScan
        Signal total power scans
    reftp:  list ~spectra.scan.TPScan
        Reference total power scans
    calibrate: bool
        Whether or not to calibrate the data.
    smoothref: int
        the number of channels in the reference to boxcar smooth prior to calibration
    apply_flags : boolean, optional.  If True, apply flags before calibration.
    observer_location : `~astropy.coordinates.EarthLocation`
        Location of the observatory. See `~dysh.coordinates.Observatory`.
        This will be transformed to `~astropy.coordinates.ITRS` using the time of
        observation DATE-OBS or MJD-OBS in
        the SDFITS header.  The default is the location of the GBT.
    weights: str
        Weighting scheme to use when averaging the signal and reference scans
        'tsys' or None.  If 'tsys' the weight will be calculated as:

         :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

        Default: 'tsys'
    """

    def __init__(
        self,
        sigtp,
        reftp,
        calibrate=True,
        smoothref=1,
        apply_flags=False,
        observer_location=Observatory["GBT"],
        **kwargs,
    ):
        ScanBase.__init__(self, sigtp[0]._sdfits)
        kwargs_opts = {
            "timeaverage": False,
            "weights": "tsys",  # or None or ndarray
            "debug": False,
        }
        kwargs_opts.update(kwargs)
        w = kwargs_opts["weights"]
        if len(reftp) != len(sigtp):
            raise ValueError(
                f"Reference and signal total power arrays are different lengths: {len(reftp)} != {len(sigtp)}"
            )
        self._scan = sigtp[0]._scan
        self._sigtp = sigtp
        self._reftp = reftp
        self._ifnum = self._sigtp[0].ifnum
        self._fdnum = self._sigtp[0].fdnum
        self._npol = self._sigtp[0].npol
        self._pols = self._sigtp[0]._pols
        self._nchan = len(reftp[0]._data[0])
        self._nrows = np.sum([stp.nrows for stp in self._sigtp])
        self._nint = 0
        self._smoothref = smoothref
        if self._smoothref > 1:
            print(f"SubBeamNodScan smoothref={self._smoothref} not implemented yet")
        self._apply_flags = apply_flags
        self._observer_location = observer_location
        self._calibrated = None
        if calibrate:
            self.calibrate(weights=w)
        self._validate_defaults()

    def calibrate(self, **kwargs):
        """Calibrate the SubBeamNodScan data"""
        nspect = len(self._reftp)
        self._tsys = np.empty(nspect, dtype=float)
        self._exposure = np.empty(nspect, dtype=float)
        self._delta_freq = np.empty(nspect, dtype=float)
        self._calibrated = np.ma.empty((nspect, self._nchan), dtype=float)

        for i in range(nspect):
            sig = self._sigtp[i].timeaverage(weights=kwargs["weights"])
            ref = self._reftp[i].timeaverage(weights=kwargs["weights"])
            # Combine sig and ref.
            ta = ((sig - ref) / ref).flux.value * ref.meta["WTTSYS"]
            self._tsys[i] = ref.meta["WTTSYS"]
            self._exposure[i] = sig.meta["EXPOSURE"]
            self._delta_freq[i] = sig.meta["CDELT1"]
            self._calibrated[i] = ta

    def calibrated(self, i):
        meta = deepcopy(self._sigtp[i].timeaverage().meta)  # use self._sigtp.meta? instead?
        meta["TSYS"] = self._tsys[i]
        meta["EXPOSURE"] = self._exposure[i]
        naxis1 = len(self._calibrated[i])
        meta["NAXIS1"] = naxis1
        if "CUNIT1" not in meta:
            meta["CUNIT1"] = "Hz"  # @todo this is in gbtfits.hdu[0].header['TUNIT11'] but is it always TUNIT11?
        meta["CUNIT2"] = "deg"  # is this always true?
        meta["CUNIT3"] = "deg"  # is this always true?
        restfrq = meta["RESTFREQ"]
        rfq = restfrq * u.Unit(meta["CUNIT1"])
        restfreq = rfq.to("Hz").value
        meta["RESTFRQ"] = restfreq  # WCS wants no E
        s = Spectrum.make_spectrum(
            Masked(self._calibrated[i] * u.K, self._calibrated[i].mask),
            meta=meta,
            observer_location=self._observer_location,
        )
        s.merge_commentary(self)
        return s

    @property
    def exposure(self):
        return self._exposure

    @property
    def delta_freq(self):
        return self._delta_freq

    def timeaverage(self, weights="tsys"):
        r"""Compute the time-averaged spectrum for this scan.

        Parameters
        ----------
        weights: str
            'tsys' or None.  If 'tsys' the weight will be calculated as:

             :math:`w = t_{exp} \times \delta\nu/T_{sys}^2`

            Default: 'tsys'
        Returns
        -------
        spectrum : :class:`~spectra.spectrum.Spectrum`
            The time-averaged spectrum

        .. note::
           Data that are masked will have values set to zero.  This is a feature of `numpy.ma.average`. Data mask fill value is NaN (np.nan)
        """
        if self._calibrated is None or len(self._calibrated) == 0:
            raise Exception("You can't time average before calibration.")
        if self._npol > 1:
            raise Exception(f"Can't yet time average multiple polarizations {self._npol}")
        self._timeaveraged = deepcopy(self.calibrated(0))
        data = self._calibrated
        if weights == "tsys":
            w = self.tsys_weight
        else:
            w = None
        self._timeaveraged._data = np.ma.average(data, axis=0, weights=w)
        self._timeaveraged._data.set_fill_value(np.nan)
        self._timeaveraged.meta["MEANTSYS"] = np.mean(self._tsys)
        self._timeaveraged.meta["WTTSYS"] = sq_weighted_avg(self._tsys, axis=0, weights=w)
        self._timeaveraged.meta["TSYS"] = self._timeaveraged.meta["WTTSYS"]
        self._timeaveraged.meta["EXPOSURE"] = np.sum(self.exposure)
        return self._timeaveraged



import argparse
import sys
import uuid
from pathlib import Path

import httpx
from rich.progress import Progress
from tenacity import retry, stop_after_attempt

from dysh.log import init_logging, logger


@retry(stop=stop_after_attempt(3))
def from_url(url, path=Path(".")):
    """
    Download a file from `url` to `path`.

    Parameters
    ----------
    url : str
        The URL of the data file
    path : str or `pathlib.Path`
        The path to the directory to save the data. If `path/filename` already exists, the file will not be downloaded again.

    Returns
    -------
    savepath : `pathlib.Path`
        The path to the downloaded (or existing) data
    """

    # Convert savepath to Path if given string
    if type(path) is str:
        path = Path(path)

    try:
        # Make the HTTPX client
        client = httpx.Client(follow_redirects=True)

        with client.stream("GET", url) as resp:

            # Get the filename from the URL
            filename = Path(resp.url.path).name

            # Check if given path is directory or filename
            if path.is_dir():
                path.mkdir(parents=True, exist_ok=True)
                savepath = path / filename
            else:
                savepath = path

            # Skip downloading if file already exists
            if savepath.exists():
                logger.info(f"{filename} already downloaded at {path}")
                return savepath

            else:
                # Download the file
                logger.info("Starting download...")
                resp.raise_for_status()

                # Download to a temporary path first
                # so desired file only shows up if successful
                tmp_path = savepath.parent / f"{filename}.{uuid.uuid4()}.tmp"

                # Write chunks to file with progress bar
                with open(tmp_path, "wb") as out_file:
                    with Progress(refresh_per_second=5) as progress:
                        task_length = int(resp.headers.get("content-length", 0))
                        task = progress.add_task("[red]Downloading...", total=task_length)
                        for chunk in resp.iter_raw():
                            out_file.write(chunk)
                            progress.update(task, advance=len(chunk))
        client.close()

    # If something goes wrong, throw Exception
    except Exception as exc:
        print(exc, file=sys.stderr)
        raise

    # Rename the temp file to the desired name
    tmp_path.rename(savepath)
    logger.info(f"Saved {filename} to {savepath}")

    return Path(savepath)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("url")
    parser.add_argument(
        "-o",
        "--output",
        dest="output_path",
        type=Path,
        help="The path where the downloaded file will be saved. If not provided, it will be derived from the URL.",
        default=Path("."),
    )

    return parser.parse_args()


def main():
    args = parse_args()
    savepath = from_url(args.url, args.output_path)
    return savepath


if __name__ == "__main__":
    init_logging(2)
    main()
